from datetime import datetime
import importlib
import inspect
import json
import logging
from multiprocessing import Event as MultiprocessingEvent, Process, Queue, set_start_method
import os
import pkgutil
from queue import Empty
import sys
import threading
import time
import traceback
from uuid import uuid4
from typing import Any, Dict, Generator, Iterator, Optional, Union, cast

from flask import Blueprint, Response, current_app as app, request, stream_with_context
import orjson
from pydantic import ValidationError
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder

from vellum.workflows import BaseInputs, BaseWorkflow
from vellum.workflows.errors import WorkflowError, WorkflowErrorCode
from vellum.workflows.events.workflow import (
    WorkflowExecutionInitiatedBody,
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionRejectedBody,
    WorkflowExecutionRejectedEvent,
)
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.nodes import BaseNode
from vellum.workflows.vellum_client import create_vellum_client
from workflow_server.config import ENABLE_PROCESS_WRAPPER, MEMORY_LIMIT_MB
from workflow_server.core.events import (
    SPAN_ID_EVENT,
    STREAM_FINISHED_EVENT,
    VEMBDA_EXECUTION_FULFILLED_EVENT_NAME,
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
    VembdaExecutionInitiatedBody,
    VembdaExecutionInitiatedEvent,
)
from workflow_server.core.executor import (
    get_random_namespace,
    stream_node_process_timeout,
    stream_workflow,
    stream_workflow_process_timeout,
)
from workflow_server.core.type_check import type_check
from workflow_server.core.utils import (
    create_vembda_rejected_event,
    is_events_emitting_enabled,
    serialize_vembda_rejected_event,
)
from workflow_server.core.workflow_executor_context import (
    DEFAULT_TIMEOUT_SECONDS,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.oom_killer import get_is_oom_killed
from workflow_server.utils.system_utils import (
    add_active_span_id,
    get_active_process_count,
    increment_process_count,
    remove_active_span_id,
    wait_for_available_process,
)
from workflow_server.utils.utils import get_version

bp = Blueprint("exec", __name__)
logger = logging.getLogger(__name__)

set_start_method("fork", force=True)

CUSTOM_NODES_DIRECTORY = "vellum_custom_nodes"
WORKFLOW_INITIATION_TIMEOUT_SECONDS = 60


@bp.route("/stream", methods=["POST"])
def stream_workflow_route() -> Response:
    data = request.get_json()
    try:
        context = WorkflowExecutorContext.model_validate(data)
    except ValidationError as e:
        error_message = e.errors()[0]["msg"]
        error_location = e.errors()[0]["loc"]

        return Response(
            json.dumps({"detail": f"Invalid context: {error_message} at {error_location}"}),
            status=400,
            content_type="application/json",
        )

    headers = _get_headers(context)

    # We can exceed the concurrency count currently with long running workflows due to a knative issue. So here
    # if we detect a memory problem just exit us early
    if not wait_for_available_process():
        return Response(
            json.dumps(
                {
                    "detail": f"Workflow server concurrent request rate exceeded. "
                    f"Process count: {get_active_process_count()}"
                }
            ),
            status=429,
            content_type="application/json",
            headers=headers,
        )

    start_workflow_state = _start_workflow(context)
    if isinstance(start_workflow_state, Response):
        return start_workflow_state

    workflow_events, vembda_initiated_event, process, span_id, headers = start_workflow_state

    def generator() -> Generator[str, None, None]:
        try:
            yield "\n"
            if not context.exclude_vembda_events:
                yield vembda_initiated_event.model_dump_json()
                yield "\n"
            for row in workflow_events:
                yield "\n"
                if isinstance(row, dict):
                    dump = orjson.dumps(row).decode("utf-8")
                    yield dump
                else:
                    yield row
                yield "\n"
            # Sometimes the connections get hung after they finish with the vembda fulfilled event
            # if it happens during a knative scale down event. So we emit an END string so that
            # we don't have to do string compares on all the events for performance.
            yield "\n"
            yield "END"
            yield "\n"

            logger.info(
                f"Workflow stream completed, execution ID: {span_id}, process count: {get_active_process_count()}"
            )
        except GeneratorExit:
            # These can happen either from Vembda disconnects (possibily from predict disconnects) or
            # from knative activator gateway timeouts which are caused by idleTimeout or responseStartSeconds
            # being exceeded.
            app.logger.warning(
                "Client disconnected in the middle of the Workflow Stream",
                extra={
                    "sentry_tags": {
                        "server_version": vembda_initiated_event.body.server_version,
                        "sdk_version": vembda_initiated_event.body.sdk_version,
                    }
                },
            )
            _emit_client_disconnect_events(
                context,
                span_id,
                "Client disconnected in the middle of the Workflow Stream",
            )
            return
        except Exception as e:
            logger.exception("Error during workflow response stream generator", extra={"error": e})
            yield "\n"
            yield "END"
            yield "\n"
            return
        finally:
            if ENABLE_PROCESS_WRAPPER:
                try:
                    if process and process.is_alive():
                        process.kill()
                    if process:
                        increment_process_count(-1)
                        remove_active_span_id(span_id)
                except Exception as e:
                    logger.error("Failed to kill process", e)
            else:
                increment_process_count(-1)
                remove_active_span_id(span_id)

    resp = Response(
        stream_with_context(generator()),
        status=200,
        content_type="application/x-ndjson",
        headers=headers,
    )
    return resp


def _emit_async_error_events(
    context: WorkflowExecutorContext, error_message: str, stacktrace: Optional[str] = None
) -> None:
    """
    Emit workflow execution error events when async execution fails before or during workflow startup.

    This ensures that errors in async mode are properly reported to Vellum's events API,
    making them visible in the executions UI.
    """
    try:
        workflow_span_id = context.workflow_span_id or str(uuid4())

        initiated_event = WorkflowExecutionInitiatedEvent[Any, Any](
            trace_id=context.trace_id,
            span_id=workflow_span_id,
            body=WorkflowExecutionInitiatedBody(inputs=BaseInputs(), workflow_definition=BaseWorkflow),
            parent=context.execution_context.parent_context if context.execution_context else None,
        )

        rejected_event = WorkflowExecutionRejectedEvent(
            trace_id=context.trace_id,
            span_id=workflow_span_id,
            body=WorkflowExecutionRejectedBody(
                error=WorkflowError(
                    message=error_message,
                    code=WorkflowErrorCode.INTERNAL_ERROR,
                ),
                stacktrace=stacktrace,
                workflow_definition=BaseWorkflow,
            ),
            parent=context.execution_context.parent_context if context.execution_context else None,
        )

        context.vellum_client.events.create(request=[initiated_event, rejected_event])  # type: ignore[list-item]
    except Exception as e:
        logger.exception(f"Failed to emit async error events: {e}")


def _emit_client_disconnect_events(
    context: WorkflowExecutorContext,
    workflow_span_id: str,
    error_message: str,
) -> None:
    """
    Emit workflow execution rejected event when a client disconnects mid-stream.

    Since the workflow has already started streaming (the initiated event was already emitted),
    we only need to emit the rejected event to properly close out the execution.
    """
    try:
        # TODO: In the future, we should capture the real workflow_definition from the initiated event
        # For now, we use BaseWorkflow as a placeholder
        rejected_event = WorkflowExecutionRejectedEvent(
            trace_id=context.trace_id,
            span_id=workflow_span_id,
            body=WorkflowExecutionRejectedBody(
                workflow_definition=BaseWorkflow,
                error=WorkflowError(
                    message=error_message,
                    code=WorkflowErrorCode.WORKFLOW_CANCELLED,
                ),
            ),
            parent=context.execution_context.parent_context if context.execution_context else None,
        )

        context.vellum_client.events.create(request=[rejected_event])  # type: ignore[list-item]
    except Exception as e:
        logger.exception(f"Failed to emit client disconnect events: {e}")


@bp.route("/async-exec", methods=["POST"])
def async_exec_workflow() -> Response:
    data = request.get_json()
    try:
        context = WorkflowExecutorContext.model_validate(data)
    except ValidationError as e:
        error_message = e.errors()[0]["msg"]
        error_location = e.errors()[0]["loc"]

        return Response(
            json.dumps({"detail": f"Invalid context: {error_message} at {error_location}"}),
            status=400,
            content_type="application/json",
        )

    # Reject back to the queue handler if were low on memory here, though maybe we should update the is_available
    # route to look at memory too. Don't send this response as an event. Though we might want some logic to catch
    # if they have a workflow server that can never start a workflow because the base image uses so much memory.
    if not wait_for_available_process():
        return Response(
            json.dumps({"detail": f"Server resources low." f"Process count: {get_active_process_count()}"}),
            status=429,
            content_type="application/json",
        )

    def run_workflow_background() -> None:
        process: Optional[Process] = None
        span_id: Optional[str] = None

        try:
            start_workflow_result = _start_workflow(context)
            if isinstance(start_workflow_result, Response):
                error_detail = start_workflow_result.get_json().get("detail", "Unknown error during workflow startup")
                _emit_async_error_events(context, error_detail)
                return

            workflow_events, vembda_initiated_event, process, span_id, headers = start_workflow_result

            for _ in workflow_events:
                # This is way inefficient in process mode since were just having the main proc stream the events
                # to nowhere wasting memory I/O and cpu.
                continue
            logger.info(
                f"Workflow async exec completed, execution ID: {span_id}, process count: {get_active_process_count()}"
            )
        except Exception as e:
            logger.exception(
                "Error during workflow async background worker",
                extra={
                    "sentry_tags": {
                        "vellum_trace_id": str(context.trace_id),
                    },
                },
            )
            _emit_async_error_events(context, str(e), traceback.format_exc())
        finally:
            if ENABLE_PROCESS_WRAPPER:
                try:
                    if process and process.is_alive():
                        process.kill()
                    if process:
                        increment_process_count(-1)
                        if span_id:
                            remove_active_span_id(span_id)
                except Exception as e:
                    logger.error("Failed to kill process", e)
            else:
                increment_process_count(-1)
                if span_id:
                    remove_active_span_id(span_id)

    thread = threading.Thread(target=run_workflow_background)
    thread.start()

    return Response(
        json.dumps({"success": True}),
        status=200,
        content_type="application/json",
    )


def _start_workflow(
    context: WorkflowExecutorContext,
) -> Union[
    Response,
    tuple[
        Iterator[Union[str, dict]],
        VembdaExecutionInitiatedEvent,
        Optional[Process],
        str,
        dict[str, str],
    ],
]:
    headers = _get_headers(context)
    logger.info(
        f"Starting Workflow Server Request, trace ID: {context.trace_id}, "
        f"process count: {get_active_process_count()}, process wrapper: {ENABLE_PROCESS_WRAPPER}"
    )

    # Create this event up here so timestamps are fully from the start to account for any unknown overhead
    version_data = get_version()
    vembda_initiated_event = VembdaExecutionInitiatedEvent(
        id=uuid4(),
        timestamp=datetime.now(),
        trace_id=context.trace_id,
        span_id=context.execution_id,
        body=VembdaExecutionInitiatedBody.model_validate(version_data),
        parent=None,
    )

    output_queue: Queue[Union[str, dict]] = Queue()
    cancel_signal = MultiprocessingEvent()
    timeout_signal = MultiprocessingEvent()

    process: Optional[Process] = None
    if ENABLE_PROCESS_WRAPPER:
        try:
            process = stream_workflow_process_timeout(
                executor_context=context,
                queue=output_queue,
                cancel_signal=cancel_signal,
                timeout_signal=timeout_signal,
            )
            increment_process_count(1)
        except Exception as e:
            logger.exception(e)

            output_queue.put(create_vembda_rejected_event(context, traceback.format_exc()))

        try:
            first_item = output_queue.get(timeout=WORKFLOW_INITIATION_TIMEOUT_SECONDS)
        except Empty:
            logger.error("Request timed out trying to initiate the Workflow")

            if process and process.is_alive():
                process.kill()
            increment_process_count(-1)

            return Response(
                json.dumps({"detail": "Request timed out trying to initiate the Workflow"}),
                status=408,
                content_type="application/json",
                headers=headers,
            )
    else:

        def workflow_stream_processor() -> Iterator[Union[dict, str]]:
            span_id_emitted = False
            try:
                workflow_iterator, span_id = stream_workflow(
                    context,
                    disable_redirect=True,
                    cancel_signal=cancel_signal,
                    timeout_signal=timeout_signal,
                )
                yield f"{SPAN_ID_EVENT}:{span_id}"
                span_id_emitted = True
                for event in workflow_iterator:
                    yield event
            except Exception as e:
                if not span_id_emitted:
                    yield f"{SPAN_ID_EVENT}:{uuid4()}"

                logger.exception(e)
                yield serialize_vembda_rejected_event(context, "Internal Server Error")

        stream_iterator = workflow_stream_processor()
        first_item = next(stream_iterator)
        increment_process_count(1)

    if not isinstance(first_item, str) or not first_item.startswith(SPAN_ID_EVENT):
        logger.error(
            "Workflow stream did not start with span id event",
            extra={
                "sentry_tags": {
                    "vellum_trace_id": str(context.trace_id),
                },
            },
        )
        return Response(
            json.dumps({"detail": "Internal Server Error"}),
            status=500,
            content_type="application/json",
            headers=headers,
        )

    span_id = first_item.split(":")[1]
    headers["X-Vellum-Workflow-Span-Id"] = span_id
    add_active_span_id(span_id)

    logger.info(f"Starting Workflow Stream, execution ID: {span_id}, ")

    def process_events(queue: Queue) -> Iterator[Union[str, dict]]:
        event: Union[str, dict]
        loops = 0
        timed_out_time: Optional[float] = None

        while True:
            loops += 1
            # Check if we timed out and kill the process if so. Set the timeout a little under what
            # the default is (30m) since the connection limit is 30m and otherwise we may not receive
            # the timeout event. After cancelling the workflow wait 5 seconds for the workflow to emit
            # any cancel events before ending the stream.
            if (
                min(context.timeout, DEFAULT_TIMEOUT_SECONDS - 90)
                < ((time.time_ns() - context.request_start_time) / 1_000_000_000)
                and not timed_out_time
            ):
                logger.error("Workflow timed out, waiting 5 seconds before ending request...")
                cancel_signal.set()
                # We pass this separate signal in so we can get the vembda time_out flag set to true in the vembda
                # fulfilled event inside of exec. In the future we might have a separate timeout event in wsdk
                timeout_signal.set()
                timed_out_time = time.time()

            if timed_out_time is not None and timed_out_time + 5 < time.time():
                logger.warning("Killing request after workflow timeout")

                if ENABLE_PROCESS_WRAPPER and process and process.is_alive():
                    process.kill()

                if not ENABLE_PROCESS_WRAPPER or process:
                    increment_process_count(-1)
                    remove_active_span_id(span_id)

                if not context.exclude_vembda_events:
                    yield VembdaExecutionFulfilledEvent(
                        id=uuid4(),
                        timestamp=datetime.now(),
                        trace_id=context.trace_id,
                        span_id=context.execution_id,
                        body=VembdaExecutionFulfilledBody(
                            exit_code=-1,
                            container_overhead_latency=context.container_overhead_latency,
                            timed_out=True,
                        ),
                        parent=None,
                    ).model_dump(mode="json")

                break

            if get_is_oom_killed():
                logger.warning("Workflow stream OOM Kill event")

                yield create_vembda_rejected_event(
                    context, f"Organization Workflow server has exceeded {MEMORY_LIMIT_MB}MB memory limit."
                )

                if process and process.is_alive():
                    process.kill()
                if process:
                    increment_process_count(-1)
                    remove_active_span_id(span_id)

                break

            try:
                if ENABLE_PROCESS_WRAPPER:
                    event = queue.get(timeout=0.1)
                else:
                    event = next(stream_iterator)
            except Empty:
                # Emit waiting event if were just sitting around to attempt to keep the line
                # open to trick knative
                if loops % 20 == 0:
                    yield "WAITING"

                    if ENABLE_PROCESS_WRAPPER and process and not process.is_alive():
                        logger.error("Workflow process exited abnormally")

                        yield create_vembda_rejected_event(
                            context, "Internal Server Error, Workflow process exited abnormally"
                        )

                        break

                continue
            except StopIteration:
                break
            except Exception as e:
                logger.exception(e)
                break

            if event == STREAM_FINISHED_EVENT:
                break
            yield event

    workflow_events = process_events(output_queue)

    return workflow_events, vembda_initiated_event, process, span_id, headers


@bp.route("/stream-node", methods=["POST"])
def stream_node_route() -> Response:
    data = request.get_json()

    try:
        context = NodeExecutorContext.model_validate(data)
    except ValidationError as e:
        error_message = e.errors()[0]["msg"]
        error_location = e.errors()[0]["loc"]
        return Response(
            json.dumps({"detail": f"Invalid context: {error_message} at {error_location}"}),
            status=400,
            content_type="application/json",
        )

    # Create this event up here so timestamps are fully from the start to account for any unknown overhead
    version_data = get_version()
    vembda_initiated_event = VembdaExecutionInitiatedEvent(
        id=uuid4(),
        timestamp=datetime.now(),
        trace_id=context.trace_id,
        span_id=context.execution_id,
        body=VembdaExecutionInitiatedBody.model_validate(version_data),
        parent=None,
    )

    app.logger.debug(f"Node stream started. Trace ID: {context.trace_id}")

    pebble_queue: Queue[dict] = Queue()
    process = stream_node_process_timeout(
        executor_context=context,
        queue=pebble_queue,
    )

    def node_events() -> Iterator[dict]:
        while True:
            try:
                event = pebble_queue.get(timeout=context.timeout)

            except Empty:
                if not process.is_alive():
                    yield create_vembda_rejected_event(context, "Internal Server Error")
                    break

            yield event
            if event.get("name") == VEMBDA_EXECUTION_FULFILLED_EVENT_NAME:
                break

    def generator() -> Generator[str, None, None]:
        yield orjson.dumps(vembda_initiated_event.model_dump(mode="json")).decode("utf-8")

        for row in node_events():
            yield "\n"
            yield orjson.dumps(row).decode("utf-8")

    headers = {
        "X-Vellum-SDK-Version": vembda_initiated_event.body.sdk_version,
        "X-Vellum-Server-Version": vembda_initiated_event.body.server_version,
        "X-Vellum-Events-Emitted": str(is_events_emitting_enabled(context)),
    }

    resp = Response(
        stream_with_context(generator()),
        status=200,
        content_type="application/x-ndjson",
        headers=headers,
    )
    return resp


@bp.route("/serialize", methods=["POST"])
def serialize_route() -> Response:
    data = request.get_json()

    request_id = uuid4()
    logger.info(
        "Starting serialize request",
        extra={
            "request_id": str(request_id),
            "X-Vellum-Copilot-Span-Id": request.headers.get("X-Vellum-Copilot-Span-Id"),
            "X-Logfire-Trace-Id": request.headers.get("X-Logfire-Trace-Id"),
            "X-Logfire-Span-Id": request.headers.get("X-Logfire-Span-Id"),
        },
    )

    files = data.get("files", {})
    workspace_api_key = data.get("workspace_api_key")
    is_new_server = data.get("is_new_server", False)
    module = data.get("module")
    ml_models = data.get("ml_models")

    # Expected: "zuban" | "mypy" | True | False.
    # Unknown (but truthy) values will use the default type checker.
    run_type_check = data.get("run_type_check", False)

    headers = {
        "X-Vellum-Is-New-Server": str(is_new_server).lower(),
    }

    if not files:
        error_message = "No files received"
        logger.warning(error_message)
        return Response(
            json.dumps({"detail": error_message}),
            status=400,
            content_type="application/json",
            headers=headers,
        )

    client = create_vellum_client(api_key=workspace_api_key)

    # Generate a unique namespace for this serialization request
    namespace = get_random_namespace()
    virtual_finder = VirtualFileFinder(files, namespace, source_module=module)

    type_check_errors = {}
    if run_type_check:
        type_check_result = type_check(files, type_checker=run_type_check)
        if not type_check_result.success:
            type_check_errors = {
                "type_errors": type_check_result.type_errors,
                "type_checker_failure": type_check_result.failure_message,
            }

    try:
        sys.meta_path.append(virtual_finder)
        result = BaseWorkflowDisplay.serialize_module(
            namespace, client=client, dry_run=True, ml_models=ml_models
        ).model_dump()

        return Response(
            json.dumps({**result, **type_check_errors}),
            status=200,
            content_type="application/json",
            headers=headers,
        )

    except WorkflowInitializationException as e:
        error_message = f"Serialization failed: {str(e)}"
        logger.warning(error_message)
        return Response(
            json.dumps({"detail": error_message, **type_check_errors}),
            status=400,
            content_type="application/json",
            headers=headers,
        )

    except Exception as e:
        logger.exception("Error during serialization")
        return Response(
            json.dumps(
                {
                    "detail": f"Serialization failed: {str(e)}",
                    "errors": [{"message": str(e)}],
                    **type_check_errors,
                }
            ),
            status=500,
            content_type="application/json",
            headers=headers,
        )

    finally:
        if virtual_finder in sys.meta_path:
            sys.meta_path.remove(virtual_finder)
        logger.info(
            "Finished serialize request",
            extra={
                "request_id": str(request_id),
            },
        )


@bp.route("/version", methods=["GET"])
def get_version_route() -> tuple[dict, int]:
    resp = get_version()

    try:
        # Discover nodes in the container
        nodes = []

        # Look for custom_nodes directory in the container
        custom_nodes_path = os.path.join(os.getcwd(), CUSTOM_NODES_DIRECTORY)
        if os.path.exists(custom_nodes_path):
            # Add the custom_nodes directory to Python path so we can import from it
            sys.path.append(os.path.dirname(custom_nodes_path))

            # Import all Python files in the custom_nodes directory
            for _, name, _ in pkgutil.iter_modules([custom_nodes_path]):
                try:
                    module = importlib.import_module(f"{CUSTOM_NODES_DIRECTORY}.{name}")
                    for _, obj in inspect.getmembers(module):
                        # Look for classes that inherit from BaseNode
                        if inspect.isclass(obj) and obj != BaseNode and issubclass(obj, BaseNode):
                            node_display_class = get_node_display_class(obj)
                            exec_config_raw = node_display_class().serialize(WorkflowDisplayContext())
                            exec_config = cast(Dict[str, Any], exec_config_raw)
                            config_module = exec_config["definition"]["module"]
                            description = (
                                exec_config["display_data"]["comment"]["value"]
                                if "comment" in exec_config["display_data"]
                                else ""
                            )
                            nodes.append(
                                {
                                    "id": str(uuid4()),
                                    "module": config_module,
                                    "name": obj.__name__,
                                    "label": exec_config["label"],
                                    "description": description,
                                    "exec_config": exec_config,
                                }
                            )
                except Exception as e:
                    logger.warning(f"Failed to load node from module {name}: {str(e)}", exc_info=True)

        resp["nodes"] = nodes
    except Exception as e:
        logger.exception(f"Failed to discover nodes: {str(e)}")
        resp["nodes"] = []

    return resp, 200


def startup_error_generator(
    vembda_initiated_event: VembdaExecutionInitiatedEvent, message: str, context: WorkflowExecutorContext
) -> Generator[str, None, None]:
    try:
        yield "\n"
        yield vembda_initiated_event.model_dump_json()
        yield "\n"
        yield serialize_vembda_rejected_event(context, message)
        yield "\n"
        yield "END"
        yield "\n"

        logger.error("Workflow stream could not start from resource constraints")
    except GeneratorExit:
        app.logger.error(
            "Client disconnected in the middle of the Startup Error Stream",
            extra={
                "sentry_tags": {
                    "server_version": vembda_initiated_event.body.server_version,
                    "sdk_version": vembda_initiated_event.body.sdk_version,
                }
            },
        )
        return


def _get_headers(context: WorkflowExecutorContext) -> dict[str, Union[str, Any]]:
    headers = {
        "X-Vellum-SDK-Version": get_version()["sdk_version"],
        "X-Vellum-Server-Version": get_version()["server_version"],
        "X-Vellum-Events-Emitted": str(is_events_emitting_enabled(context)),
    }
    return headers



import pathlib
import tempfile
from typing import Any, Literal, Union

from workflow_server.core.type_checker import get_checker
from workflow_server.core.type_checker.base import TypeCheckResult


def type_check(
    files: dict[str, str],
    type_checker: Union[Literal["zuban", "mypy"], None, Any] = None,
) -> TypeCheckResult:
    if type_checker in ("mypy", "zuban"):
        runner = get_checker(type_checker)
    else:
        runner = get_checker()

    configuration_status = runner.is_configured()
    if not configuration_status[0]:
        return TypeCheckResult(
            success=False,
            failure_message=configuration_status[1],
        )

    with tempfile.TemporaryDirectory(prefix="pws_type_check_") as tmpdir:
        tmpdir_path = pathlib.Path(tmpdir).resolve()
        write_files(tmpdir_path, files)

        return runner.run(tmpdir_path)


def write_files(tmpdir_path: pathlib.Path, files: dict[str, str]) -> None:
    for file_name, file_content in files.items():
        # Skip directory entries
        if file_name.endswith("/"):
            continue

        raw_path = pathlib.Path(file_name)

        # Users shouldn't specify absolute paths
        # since we stuff files into random namespaces.
        if raw_path.is_absolute():
            raise ValueError(f"File {raw_path} is absolute")

        target = (tmpdir_path / raw_path).resolve()
        if not target.is_relative_to(tmpdir_path):
            raise ValueError(f"File {raw_path} is not in {tmpdir_path}")

        # Create the parent directories if they don't exist
        target.parent.mkdir(parents=True, exist_ok=True)

        # Write the file content
        target.write_text(file_content)



from abc import ABC, abstractmethod
import dataclasses
import pathlib
from typing import Literal, Optional, Tuple, Union

ConfigurationStatus = Union[
    Tuple[Literal[True], None],
    Tuple[Literal[False], str],
]


@dataclasses.dataclass
class TypeCheckResult:
    success: bool

    type_errors: Optional[str] = None
    failure_message: Optional[str] = None


class WorkflowTypeChecker(ABC):
    def __init__(self) -> None:
        self._is_configured = False

    @abstractmethod
    def is_configured(self) -> ConfigurationStatus:
        pass

    def run(self, dir: pathlib.Path) -> TypeCheckResult:
        if not self._is_configured:
            return TypeCheckResult(
                success=False,
                failure_message="Type checker is not configured",
            )

        res = self._run(dir)
        if res.type_errors:
            res.type_errors = self._clean_type_errors(res.type_errors, dir)

        return res

    @abstractmethod
    def _run(self, dir: pathlib.Path) -> TypeCheckResult:
        pass

    @abstractmethod
    def _clean_type_errors(self, type_errors: str, dir: pathlib.Path) -> str:
        """
        Clean temp directory from type checker output.

        Implementation may be specific to each type checker, depending
        on the working directory used by the type checker.
        """
        pass



import pytest

from deepdiff import DeepDiff

from workflow_server.server import create_app


@pytest.mark.parametrize(
    "type_checker",
    [
        True,
        "zuban",
        "mypy",
    ],
)
def test_serialize_route__with__workflow_type_error(type_checker):
    # GIVEN a Flask application
    flask_app = create_app()

    # AND a complete workflow with multiple files
    workflow_files = {
        "__init__.py": """\
# flake8: noqa: F401, F403
""",
        "inputs.py": """\
from typing import Any, Optional

from vellum.workflows.inputs import BaseInputs


class Inputs(BaseInputs):
    text: str
    var_1: Optional[Any]
""",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .inputs import Inputs
from .nodes.final_output import FinalOutput
from .nodes.my_custom_node import MyCustomNode


class Workflow(BaseWorkflow[Inputs, BaseState]):
    graph = MyCustomNode >> FinalOutput

    class Outputs(BaseWorkflow.Outputs):
        final_output = FinalOutput.Outputs.value

""",
        "nodes/__init__.py": """\
from .final_output import FinalOutput
from .my_custom_node import MyCustomNode

__all__ = [
    "FinalOutput",
    "MyCustomNode",
]
""",
        "nodes/my_custom_node.py": """\
from vellum.workflows.nodes import BaseNode


class MyCustomNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> Outputs:
        schema_response = (
            self._context.vellum_client.integrations.execute_integration_tool(
                "test_integration_name",
                "test_integration_provider",
                "test_tool_name",
                arguments={
                    "text": "some_text",
                },
                bad_kwarg=None,  # ERROR: Bad kwarg.
            )
        )
        _ = schema_response.output  # ERROR: Invalid attribute.

        return self.Outputs(value=schema_response.provider)
""",
        "nodes/final_output.py": """\
from vellum.workflows.nodes.displayable import FinalOutputNode
from vellum.workflows.state import BaseState

from .my_custom_node import MyCustomNode


class FinalOutput(FinalOutputNode[BaseState, str]):
    class Outputs(FinalOutputNode.Outputs):
        value = MyCustomNode.Outputs.value
""",
    }

    # WHEN we make a request to the serialize route
    with flask_app.test_client() as test_client:
        response = test_client.post(
            "/workflow/serialize",
            json={"files": workflow_files, "run_type_check": type_checker},
        )

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the response should contain the full WorkflowSerializationResult
    assert "exec_config" in response.json
    assert "errors" in response.json
    assert "dataset" in response.json or response.json.get("dataset") is None
    exec_config = response.json["exec_config"]

    # AND the exec_config should have workflow_raw_data
    assert "workflow_raw_data" in exec_config
    nodes = exec_config["workflow_raw_data"]["nodes"]

    # AND we should find the workflow nodes
    node_labels = {node["data"]["label"] if "data" in node else node["label"] for node in nodes}
    expected_nodes = {"My Custom Node", "Final Output", "Entrypoint Node"}

    assert not DeepDiff(node_labels, expected_nodes, ignore_order=True)

    # AND we should see certain type errors
    expected_type_errors = [
        'nodes/my_custom_node.py:10: error: Unexpected keyword argument "bad_kwarg" for "execute_integration_tool" of "IntegrationsClient"  [call-arg]',  # noqa: E501
        'nodes/my_custom_node.py:20: error: "ComposioExecuteToolResponse" has no attribute "output"  [attr-defined]',  # noqa: E501
    ]
    actual_type_errors = response.json["type_errors"].splitlines()

    assert all(
        expected_type_error in actual_type_errors for expected_type_error in expected_type_errors
    ), actual_type_errors


@pytest.mark.parametrize(
    "type_checker",
    [
        True,
        "zuban",
        "mypy",
    ],
)
def test_type_check__empty_directory(type_checker):
    # Ensure we don't write empty directories as files.

    # GIVEN a Flask application
    flask_app = create_app()

    # AND a complete workflow with multiple files
    workflow_files = {
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from vellum.workflows.inputs import BaseInputs

from typing import Optional, Any

class Inputs(BaseInputs):
    text: str
    var_1: Optional[Any]

class Workflow(BaseWorkflow[Inputs, BaseState]):
    pass
""",
        "nodes/execute_code_node/": "",
        "nodes/execute_code_node/scripts/": "",
        "nodes/execute_fixed_code_node/": "",
        "nodes/execute_fixed_code_node/scripts/": "",
        "nodes/output.py": "a: int = '1'",
    }

    # WHEN we make a request to the serialize route
    with flask_app.test_client() as test_client:
        response = test_client.post(
            "/workflow/serialize",
            json={"files": workflow_files, "run_type_check": type_checker},
        )

    # THEN we should get a successful response
    assert response.status_code == 200, response.json

    # AND we should see certain type errors
    actual_type_errors = response.json["type_errors"]
    assert (
        actual_type_errors
        == """\
nodes/output.py:1: error: Incompatible types in assignment (expression has type "str", variable has type "int")  [assignment]
Found 1 error in 1 file (checked 2 source files)
"""
    ), actual_type_errors



from unittest.mock import MagicMock, patch
from uuid import uuid4
from typing import Any

from vellum.workflows.context import ExecutionContext
from vellum.workflows.errors import WorkflowErrorCode
from workflow_server.api.workflow_view import _emit_async_error_events
from workflow_server.core.workflow_executor_context import WorkflowExecutorContext


def _create_test_context(
    workflow_span_id: Any = None,
    execution_context: Any = None,
) -> WorkflowExecutorContext:
    """Helper to create a WorkflowExecutorContext for testing."""
    return WorkflowExecutorContext(
        execution_id=uuid4(),
        files={"__init__.py": "", "workflow.py": ""},
        environment_api_key="test-api-key",
        module="workflow",
        workflow_span_id=workflow_span_id,
        execution_context=execution_context or ExecutionContext(),
    )


def test_emit_async_error_events__happy_path():
    """
    Tests that _emit_async_error_events emits both initiated and rejected events to the Vellum API.
    """
    # GIVEN a workflow context with a workflow_span_id
    workflow_span_id = uuid4()
    context = _create_test_context(workflow_span_id=workflow_span_id)

    # AND a mock vellum client
    mock_events_create = MagicMock()

    # WHEN we call _emit_async_error_events
    with patch.object(context, "vellum_client") as mock_client:
        mock_client.events.create = mock_events_create
        _emit_async_error_events(context, "Test error message")

    # THEN the events.create method should be called once
    mock_events_create.assert_called_once()

    # AND it should be called with a list of two events
    call_args = mock_events_create.call_args
    events = call_args.kwargs["request"]
    assert len(events) == 2

    # AND the first event should be a WorkflowExecutionInitiatedEvent
    initiated_event = events[0]
    assert initiated_event.name == "workflow.execution.initiated"
    assert initiated_event.trace_id == context.trace_id
    assert str(initiated_event.span_id) == str(workflow_span_id)

    # AND the second event should be a WorkflowExecutionRejectedEvent with the error message
    rejected_event = events[1]
    assert rejected_event.name == "workflow.execution.rejected"
    assert rejected_event.trace_id == context.trace_id
    assert str(rejected_event.span_id) == str(workflow_span_id)
    assert rejected_event.body.error.message == "Test error message"
    assert rejected_event.body.error.code == WorkflowErrorCode.INTERNAL_ERROR


def test_emit_async_error_events__logs_exception_on_failure(caplog):
    """
    Tests that _emit_async_error_events logs an exception when the API call fails.
    """
    # GIVEN a workflow context
    context = _create_test_context(workflow_span_id=uuid4())

    # AND a mock vellum client that raises an exception
    mock_events_create = MagicMock(side_effect=Exception("API connection failed"))

    # WHEN we call _emit_async_error_events
    with patch.object(context, "vellum_client") as mock_client:
        mock_client.events.create = mock_events_create
        _emit_async_error_events(context, "Test error message")

    # THEN the function should not raise an exception
    # (implicit - if we get here, no exception was raised)

    # AND the exception should be logged
    assert any("Failed to emit async error events" in record.message for record in caplog.records)



import gc
from importlib.metadata import version
import re
import sys
from typing import Any, List

from vellum import (
    ArrayInput,
    ArrayVellumValue,
    ChatHistoryInput,
    ErrorInput,
    FunctionCallInput,
    SearchResultsInput,
    VellumAudio,
    VellumDocument,
    VellumImage,
    VellumValue,
    VellumVideo,
)
from workflow_server.config import CONTAINER_IMAGE, is_development


def unwrap_vellum_value(item: VellumValue) -> Any:
    """Recursively unwrap VellumValue objects to their primitive values.

    This is needed because ArrayInput.value returns List[VellumValue] objects,
    but workflows expect primitive Python values (str, int, etc.).
    """
    if isinstance(item, ArrayVellumValue):
        if item.value is None:
            return None
        return [unwrap_vellum_value(nested_item) for nested_item in item.value]
    else:
        return item.value


def convert_json_inputs_to_vellum(inputs: List[dict]) -> dict:
    vellum_inputs: dict[str, Any] = {}

    for input in inputs:
        value = input["value"]
        # sync with vellum-python-sdks/ee/codegen/src/context/input-variable-context.ts
        name = to_valid_python_identifier(input["name"], "input")

        # If a value is null, we don't need to bother with any validation or deserialization
        if value is None:
            vellum_inputs[name] = None
            continue

        type = input["type"]
        if type == "CHAT_HISTORY":
            vellum_inputs[name] = ChatHistoryInput.model_validate(input).value
        elif type == "FUNCTION_CALL":
            vellum_inputs[name] = FunctionCallInput.model_validate(input).value
        elif type == "SEARCH_RESULTS":
            vellum_inputs[name] = SearchResultsInput.model_validate(input).value
        elif type == "ERROR":
            vellum_inputs[name] = ErrorInput.model_validate(input).value
        elif type == "ARRAY":
            array_value = ArrayInput.model_validate(input).value
            vellum_inputs[name] = [unwrap_vellum_value(item) for item in array_value]
        # Once we export *Input classes for these two cases, we can add the union to the WorkflowExecutorContext
        # model and simplify this method to just a {to_python_safe_snake_case(input.name): input.value} mapping
        elif type == "IMAGE":
            vellum_inputs[name] = VellumImage.model_validate(value)
        elif type == "AUDIO":
            vellum_inputs[name] = VellumAudio.model_validate(value)
        elif type == "VIDEO":
            vellum_inputs[name] = VellumVideo.model_validate(value)
        elif type == "DOCUMENT":
            vellum_inputs[name] = VellumDocument.model_validate(value)
        else:
            vellum_inputs[name] = value

    return vellum_inputs


def get_version() -> dict:
    # Return hotswappable lock file so we can save it and reuse it
    lock_file = None
    try:
        with open("/app/uv.lock", "r") as f:
            lock_file = f.read()
    except Exception:
        pass

    return {
        "sdk_version": version("vellum-ai"),
        "server_version": "local" if is_development() else version("vellum-workflow-server"),
        "container_image": CONTAINER_IMAGE,
        "lock_file": lock_file,
    }


def to_python_safe_snake_case(string: str, safety_prefix: str = "_") -> str:
    # Strip special characters from start of string
    cleaned_str = re.sub(r"^[^a-zA-Z0-9_]+", "", string)

    # Check if cleaned string starts with a number or an underscore
    starts_with_unsafe = bool(re.match(r"^[\d_]", cleaned_str))

    # Convert to snake case
    snake_case = re.sub(r"([a-z])([A-Z])", r"\1_\2", cleaned_str)  # Insert underscore between lower and upper case
    snake_case = re.sub(r"[^a-zA-Z0-9]+", "_", snake_case)  # Replace any non-alphanumeric chars with underscore
    snake_case = re.sub(r"^_+|_+$", "", snake_case)  # Remove leading/trailing underscores
    snake_case = snake_case.lower()

    # Add safety prefix if needed
    cleaned_safety_prefix = (
        "_" if safety_prefix == "_" else f"{safety_prefix}{'_' if not safety_prefix.endswith('_') else ''}"
    )
    return f"{cleaned_safety_prefix}{snake_case}" if starts_with_unsafe else snake_case


def to_valid_python_identifier(string: str, safety_prefix: str = "_") -> str:
    # Strip special characters from start of string
    cleaned_str = re.sub(r"^[^a-zA-Z0-9_]+", "", string)

    # Check if cleaned string starts with a number or an underscore
    starts_with_unsafe = bool(re.match(r"^[\d_]", cleaned_str))

    # Check if the string is already a valid Python identifier (preserve case)
    is_valid_python_identifier = bool(re.match(r"^[a-zA-Z][a-zA-Z0-9]*$", cleaned_str))

    if is_valid_python_identifier and not starts_with_unsafe:
        return cleaned_str

    return to_python_safe_snake_case(string, safety_prefix)


def get_obj_size(obj: Any) -> int:
    marked = {id(obj)}
    obj_q = [obj]
    sz = 0

    while obj_q:
        sz += sum(map(sys.getsizeof, obj_q))

        # Lookup all the object referred to by the object in obj_q.
        # See: https://docs.python.org/3.7/library/gc.html#gc.get_referents
        all_refr = ((id(o), o) for o in gc.get_referents(*obj_q))

        # Filter object that are already marked.
        # Using dict notation will prevent repeated objects.
        new_refr = {o_id: o for o_id, o in all_refr if o_id not in marked and not isinstance(o, type)}

        # The new obj_q will be the ones that were not marked,
        # and we will update marked with their ids so we will
        # not traverse them again.
        obj_q = new_refr.values()  # type: ignore
        marked.update(new_refr.keys())

    return sz



import pathlib
import re
import shutil
import subprocess
from typing import Optional

from workflow_server.core.type_checker.base import ConfigurationStatus, TypeCheckResult, WorkflowTypeChecker


class ZubanWorkflowTypeChecker(WorkflowTypeChecker):
    def __init__(self, zuban_path: Optional[str] = None):
        super().__init__()

        if not zuban_path:
            zuban_path = shutil.which("zuban")

        self.zuban_path = zuban_path

    def is_configured(self) -> ConfigurationStatus:
        if not self.zuban_path:
            return False, "zuban not found"

        self._is_configured = True
        return True, None

    def _run(self, dir: pathlib.Path) -> TypeCheckResult:
        if not self.zuban_path:
            return TypeCheckResult(success=False, failure_message="zuban not found")

        try:
            data_path = pathlib.Path(__file__).parent / "serialization.mypy.ini"
            res = subprocess.run(
                [self.zuban_path, "check", dir.name, "--config-file", str(data_path)],
                capture_output=True,
                text=True,
                cwd=dir.parent,
            )
        except OSError as e:
            return TypeCheckResult(
                success=False,
                failure_message=f"Unexpected failure running zuban: {e}",
            )

        if res.returncode == 0:
            return TypeCheckResult(
                success=True,
            )
        elif res.returncode == 1:
            return TypeCheckResult(
                success=False,
                type_errors=res.stdout,
                failure_message=res.stderr,
            )
        else:
            return TypeCheckResult(
                success=False,
                failure_message=f"Unexpected failure running zuban:\n{res.stderr}",
            )

    def _clean_type_errors(self, type_errors: str, dir: pathlib.Path) -> str:
        dir_prefix = re.escape(str(dir.name))
        pattern = re.compile(f"{dir_prefix}/?")
        return pattern.sub("", type_errors)



import json
from typing import Any, Dict

from flask import Flask, Request, Response
import jwt
from jwt import ExpiredSignatureError

from workflow_server.config import IS_ASYNC_MODE, IS_VPC, NAMESPACE, VEMBDA_PUBLIC_KEY, is_development


class AuthMiddleware:
    def __init__(self, app: Flask) -> None:
        self.app = app

    def __call__(self, environ: Dict[str, Any], start_response: Any) -> Any:
        try:
            request = Request(environ)
            if not request.path.startswith("/healthz") and not is_development() and not IS_VPC and not IS_ASYNC_MODE:
                token = request.headers.get("X-Vembda-Signature")
                if token:
                    decoded = jwt.decode(token, VEMBDA_PUBLIC_KEY, algorithms=["RS256"])

                    # Validate that the signature's namespace matches our services namespace. We need to do
                    # this because otherwise you could simply use your signature on someone else's service
                    # if you manage to get past the networking policies.
                    token_namespace = decoded.get("namespace")

                    if decoded.get("namespace") != NAMESPACE:
                        res = Response(
                            json.dumps(
                                {
                                    "detail": f"Token namespace '{token_namespace}' "
                                    f"does not match service namespace '{NAMESPACE}'"
                                }
                            ),
                            mimetype="application/json",
                            status=401,
                        )
                        return res(environ, start_response)
                else:
                    res = Response(json.dumps({"detail": "Signature missing"}), mimetype="application/json", status=401)
                    return res(environ, start_response)

        except ExpiredSignatureError:
            res = Response(
                json.dumps({"detail": "Signature token has expired. Please obtain a new token."}),
                mimetype="application/json",
                status=401,
            )
            return res(environ, start_response)
        except Exception as e:
            res = Response(
                json.dumps({"detail": f"Invalid signature token {str(e)}"}), mimetype="application/json", status=401
            )
            return res(environ, start_response)

        return self.app(environ, start_response)



from typing import Tuple

from flask import Blueprint, Response, jsonify

from workflow_server.config import CONCURRENCY
from workflow_server.utils.system_utils import get_active_process_count

bp = Blueprint("status", __name__)


@bp.route("/is_available", methods=["GET"])
def is_available() -> Tuple[Response, int]:
    resp = jsonify(
        available=get_active_process_count() < CONCURRENCY,
        process_count=get_active_process_count(),
        max_concurrency=CONCURRENCY,
    )

    return resp, 200



from typing import Literal

from workflow_server.core.type_checker.base import WorkflowTypeChecker
from workflow_server.core.type_checker.mypy import MypyWorkflowTypeChecker
from workflow_server.core.type_checker.zuban import ZubanWorkflowTypeChecker

__all__ = ["get_checker"]


def get_checker(
    type_checker: Literal["zuban", "mypy"] = "zuban",
) -> WorkflowTypeChecker:
    if type_checker == "zuban":
        return ZubanWorkflowTypeChecker()
    elif type_checker == "mypy":
        return MypyWorkflowTypeChecker()
    else:
        raise NotImplementedError(f"Unsupported type checker: {type_checker}")



import pytest
from contextlib import redirect_stdout
from importlib.metadata import version
import io
import json
from queue import Empty
import re
import time
from unittest import mock
from uuid import uuid4

from pydantic_core import PydanticSerializationError
import requests_mock

from vellum.workflows.emitters.base import WorkflowEvent
from vellum.workflows.emitters.vellum_emitter import VellumEmitter
from workflow_server.code_exec_runner import run_code_exec_stream
from workflow_server.core.executor import _dump_event
from workflow_server.server import create_app
from workflow_server.utils.system_utils import get_active_process_count


def flask_stream(request_body: dict) -> tuple[int, list]:
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

        return status_code, [
            json.loads(line)
            for line in response.data.decode().split("\n")
            if line
            and line
            not in [
                "WAITING",
                "END",
            ]
        ]


@mock.patch("workflow_server.api.workflow_view.ENABLE_PROCESS_WRAPPER", False)
def flask_stream_disable_process_wrapper(request_body: dict) -> tuple[int, list]:
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

        return status_code, [
            json.loads(line)
            for line in response.data.decode().split("\n")
            if line
            and line
            not in [
                "WAITING",
                "END",
            ]
        ]


def code_exec_stream(request_body: dict) -> tuple[int, list]:
    output = io.StringIO()

    with mock.patch("os.read") as mocked_os_read, redirect_stdout(output):
        mocked_os_read.return_value = (json.dumps(request_body) + "\n--vellum-input-stop--\n").encode("utf-8")
        run_code_exec_stream()

    lines = output.getvalue().split("\n")
    events = []
    for line in lines:
        if "--event--" in line:
            events.append(json.loads(line.replace("--event--", "")))

    return 200, events


@pytest.fixture(params=[flask_stream, code_exec_stream, flask_stream_disable_process_wrapper])
def both_stream_types(request):
    return request.param


def test_stream_workflow_flask_route__verify_headers():
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    pass
""",
        },
    }

    # WHEN we call the stream route
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

    # THEN we get a 200 response
    assert status_code == 200, response.text

    # AND the version headers are present
    assert "X-Vellum-SDK-Version" in response.headers
    assert "X-Vellum-Server-Version" in response.headers
    assert "X-Vellum-Workflow-Span-Id" in response.headers


def test_stream_workflow_route__happy_path(both_stream_types):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "test",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    with mock.patch("builtins.open", mock.mock_open(read_data="104857600")):
        # WHEN we call the stream route
        ts_ns = time.time_ns()
        request_body["vembda_service_initiated_timestamp"] = ts_ns
        status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert events[1]["body"]["workflow_definition"]["module"] == ["test", "workflow"]
    assert "display_context" in events[1]["body"], events[1]["body"]
    display_context = events[1]["body"]["display_context"]
    assert "node_displays" in display_context
    assert "workflow_inputs" in display_context
    assert "workflow_outputs" in display_context
    assert isinstance(display_context["node_displays"], dict)
    assert isinstance(display_context["workflow_inputs"], dict)
    assert isinstance(display_context["workflow_outputs"], dict)
    assert "foo" in display_context["workflow_outputs"]

    # AND the initiated event should have server_metadata with version info and memory usage
    assert "server_metadata" in events[1]["body"], events[1]["body"]
    server_metadata = events[1]["body"]["server_metadata"]
    assert server_metadata is not None, "server_metadata should not be None"
    assert "server_version" in server_metadata
    assert "sdk_version" in server_metadata
    assert "memory_usage_mb" in server_metadata
    assert isinstance(server_metadata["memory_usage_mb"], (int, float))
    assert "is_new_server" in server_metadata
    assert server_metadata["is_new_server"] is False

    # AND the initiated event should have initiated_latency within a reasonable range
    assert "initiated_latency" in server_metadata, "initiated_latency should be present in server_metadata"
    initiated_latency = server_metadata["initiated_latency"]
    assert isinstance(initiated_latency, int), "initiated_latency should be an integer (nanoseconds)"
    # Latency should be positive and less than 60 seconds (60_000_000_000 nanoseconds) for CI
    assert (
        0 < initiated_latency < 60_000_000_000
    ), f"initiated_latency should be between 0 and 60 seconds, got {initiated_latency} ns"

    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]
    assert events[2]["body"]["workflow_definition"]["module"] == ["test", "workflow"]

    # AND the fulfilled event should have server_metadata with memory usage
    assert "server_metadata" in events[2]["body"], events[2]["body"]
    fulfilled_metadata = events[2]["body"]["server_metadata"]
    assert fulfilled_metadata is not None, "fulfilled server_metadata should not be None"
    assert "memory_usage_mb" in fulfilled_metadata
    assert isinstance(fulfilled_metadata["memory_usage_mb"], (int, float))

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__happy_path_with_inputs(both_stream_types):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [
            {"name": "foo", "type": "STRING", "value": "hello"},
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    foo: str
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert "display_context" in events[1]["body"], events[1]["body"]
    display_context = events[1]["body"]["display_context"]
    assert "node_displays" in display_context
    assert "workflow_inputs" in display_context
    assert "workflow_outputs" in display_context
    assert isinstance(display_context["node_displays"], dict)
    assert isinstance(display_context["workflow_inputs"], dict)
    assert isinstance(display_context["workflow_outputs"], dict)
    assert "foo" in display_context["workflow_inputs"]
    assert "foo" in display_context["workflow_outputs"]
    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__happy_path_with_state(both_stream_types):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "state": {"foo": "bar"},
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from .state import State

class Workflow(BaseWorkflow[BaseInputs, State]):
    class Outputs(BaseWorkflow.Outputs):
        foo = State.foo
""",
            "state.py": """\
from vellum.workflows.state import BaseState

class State(BaseState):
    foo: str
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert "display_context" in events[1]["body"], events[1]["body"]
    display_context = events[1]["body"]["display_context"]
    assert "node_displays" in display_context
    assert "workflow_inputs" in display_context
    assert "workflow_outputs" in display_context
    assert isinstance(display_context["node_displays"], dict)
    assert isinstance(display_context["workflow_inputs"], dict)
    assert isinstance(display_context["workflow_outputs"], dict)
    assert "foo" in display_context["workflow_outputs"]
    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]
    assert events[2]["body"]["outputs"] == {"foo": "bar"}

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__bad_indent_in_inputs_file(both_stream_types):
    # GIVEN a valid request body
    span_id = uuid4()
    trace_id = uuid4()
    parent_span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "execution_context": {
            "trace_id": str(trace_id),
            "parent_context": {"span_id": str(parent_span_id)},
        },
        "inputs": [
            {"name": "foo", "type": "STRING", "value": "hello"},
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

  class Inputs(BaseInputs):
     foo: str
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events: vembda initiated, workflow initiated, workflow rejected, vembda fulfilled
    assert len(events) == 4

    assert events[0] == {
        "id": mock.ANY,
        "trace_id": str(trace_id),
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated"
    assert events[1]["trace_id"] == str(trace_id), "workflow initiated event should use request trace_id"
    assert events[1]["parent"] is not None, "workflow initiated event should have parent context"
    assert events[1]["parent"]["span_id"] == str(
        parent_span_id
    ), "workflow initiated event parent should match request parent_context"

    assert events[2]["name"] == "workflow.execution.rejected"
    assert events[2]["trace_id"] == str(trace_id), "workflow rejected event should use request trace_id"
    assert events[2]["span_id"] == events[1]["span_id"]
    assert events[2]["parent"] is not None, "workflow rejected event should have parent context"
    assert events[2]["parent"]["span_id"] == str(
        parent_span_id
    ), "workflow rejected event parent should match request parent_context"
    assert (
        "Syntax Error raised while loading Workflow: "
        "unexpected indent (inputs.py, line 3)" in events[2]["body"]["error"]["message"]
    )

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }


def test_stream_workflow_route__invalid_inputs_initialization_events(both_stream_types):
    """
    Tests that invalid inputs initialization gets us back a workflow initiated and workflow rejected event.
    """
    # GIVEN a valid request body with valid inputs file but omitting required input to cause
    # WorkflowInitializationException
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [
            # Omit the required input to trigger WorkflowInitializationException
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    foo: str
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events: vembda initiated, workflow initiated, workflow rejected, vembda fulfilled
    assert len(events) == 4

    # AND the first event should be vembda execution initiated
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[0]["span_id"] == str(span_id)

    # AND the second event should be workflow execution initiated
    assert events[1]["name"] == "workflow.execution.initiated"

    # AND the third event should be workflow execution rejected
    assert events[2]["name"] == "workflow.execution.rejected"
    assert events[1]["span_id"] == events[2]["span_id"]
    actual_error_message = events[2]["body"]["error"]["message"]
    assert "Required input variables" in actual_error_message
    assert "foo" in actual_error_message
    assert "should have defined value" in actual_error_message

    # AND the fourth event should be vembda execution fulfilled
    assert events[3]["name"] == "vembda.execution.fulfilled"
    assert events[3]["span_id"] == str(span_id)
    assert events[3]["body"]["exit_code"] == 0


@pytest.mark.parametrize(
    ["execute_workflow_stream", "assert_last_request"],
    [
        (flask_stream, False),  # Unfortunately, can't make assertions on requests made in a subprocess.
        (code_exec_stream, True),
        (flask_stream_disable_process_wrapper, True),
    ],
    ids=["flask_stream", "code_exec_stream", "flask_stream_disable_process_wrapper"],
)
def test_stream_workflow_route__cancel(execute_workflow_stream, assert_last_request):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "vembda_public_url": "http://test.biz",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
import time

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.workflows.base import BaseWorkflow


class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> Outputs:
        time.sleep(2)
        return self.Outputs(value="hello world")


class BasicCancellableWorkflow(BaseWorkflow):
    graph = StartNode
    class Outputs(BaseWorkflow.Outputs):
        final_value = StartNode.Outputs.value

""",
        },
    }

    # WHEN we call the stream route with a mock cancelled return true
    with requests_mock.Mocker() as mocker:
        response_mock = mocker.get(
            re.compile("http://test.biz/vembda-public/cancel-workflow-execution-status"), json={"cancelled": True}
        )
        status_code, events = execute_workflow_stream(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected cancelled events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert "display_context" in events[1]["body"], events[1]["body"]

    cancelled_event = events[-2]
    assert cancelled_event["name"] == "workflow.execution.rejected"
    assert cancelled_event["body"]["error"]["message"] == "Workflow run cancelled"

    # AND we called the cancel endpoint with the correct execution id
    workflow_span_id = events[1]["span_id"]
    if assert_last_request:
        assert response_mock.last_request
        assert (
            response_mock.last_request.url
            == f"http://test.biz/vembda-public/cancel-workflow-execution-status/{workflow_span_id}"
        )


def test_stream_workflow_route__timeout_emits_rejection_events():
    """
    Tests that when a workflow times out, we emit node and workflow rejection events.
    """

    span_id = uuid4()
    request_body = {
        "timeout": 1,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
import time

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.workflows.base import BaseWorkflow


class LongRunningNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> Outputs:
        time.sleep(30)
        return self.Outputs(value="hello world")


class TimeoutWorkflow(BaseWorkflow):
    graph = LongRunningNode
    class Outputs(BaseWorkflow.Outputs):
        final_value = LongRunningNode.Outputs.value

""",
        },
    }

    status_code, events = flask_stream(request_body)

    assert status_code == 200

    event_names = [e["name"] for e in events]

    assert "vembda.execution.initiated" in event_names
    assert "workflow.execution.initiated" in event_names
    assert "node.execution.initiated" in event_names

    assert "node.execution.rejected" in event_names, "Should emit node.execution.rejected on timeout"
    node_execution_rejected = next(e for e in events if e["name"] == "node.execution.rejected")
    assert "vellum/workflows/runner/runner.py" in node_execution_rejected["body"]["stacktrace"]

    assert "workflow.execution.rejected" in event_names, "Should emit workflow.execution.rejected on timeout"
    workflow_execution_rejected = next(e for e in events if e["name"] == "workflow.execution.rejected")
    assert workflow_execution_rejected["body"]["error"]["code"] == "WORKFLOW_TIMEOUT"
    # TODO: Uncomment once version 1.8.1 is released
    # assert "stacktrace" in workflow_execution_rejected["body"]
    # assert "vellum/workflows/runner/runner.py" in workflow_execution_rejected["body"]["stacktrace"]

    assert "vembda.execution.fulfilled" in event_names
    vembda_fulfilled = next(e for e in events if e["name"] == "vembda.execution.fulfilled")
    assert vembda_fulfilled["body"]["timed_out"] is True


def test_stream_workflow_route__very_large_events(both_stream_types):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [
            {"name": "foo", "type": "STRING", "value": "hello" * 10_000_000},
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    foo: str
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert "display_context" in events[1]["body"], events[1]["body"]
    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "links": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__happy_path_run_from_node(both_stream_types):
    # GIVEN a valid request body representing a run from a node
    node_id = uuid4()
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "node_id": str(node_id),
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "from .display import *",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.start import StartNode
from .nodes.end import EndNode

class Workflow(BaseWorkflow):
    graph = StartNode >> EndNode

    class Outputs(BaseWorkflow.Outputs):
        foo = EndNode.Outputs.value
""",
            "nodes/__init__.py": """
from .start import StartNode
from .end import EndNode

__all__ = ["StartNode", "EndNode"]
""",
            "nodes/start.py": """\
from vellum.workflows.nodes import BaseNode

class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = "apple"
""",
            "nodes/end.py": """\
from vellum.workflows.nodes import BaseNode
from .start import StartNode

class EndNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = StartNode.Outputs.value.coalesce("banana")
""",
            "display/__init__.py": """
from .nodes import *
from .workflow import *
""",
            "display/workflow.py": """
from uuid import UUID
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from vellum_ee.workflows.display.base import EdgeDisplay
from ..nodes.start import StartNode
from ..nodes.end import EndNode

class WorkflowDisplay(BaseWorkflowDisplay):
    edge_displays = {
        (StartNode.Ports.default, EndNode): EdgeDisplay(
            id=UUID("63606ff1-1c70-4516-92c6-cbaba9336424")
        ),
    }
""",
            "display/nodes/__init__.py": """
from .start import StartNodeDisplay
from .end import EndNodeDisplay

__all__ = ["StartNodeDisplay", "EndNodeDisplay"]
""",
            "display/nodes/start.py": """\
from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from ...nodes.start import StartNode

class StartNodeDisplay(BaseNodeDisplay[StartNode]):
    pass
""",
            "display/nodes/end.py": f"""\
from uuid import UUID

from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from ...nodes.end import EndNode

class EndNodeDisplay(BaseNodeDisplay[EndNode]):
    node_id = UUID("{node_id}")
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected workflow and node fulfilled events
    assert events[-4]["name"] == "node.execution.fulfilled", json.dumps(events[-1])
    assert events[-4]["body"]["node_definition"]["id"] == str(node_id)
    assert events[-2]["name"] == "workflow.execution.fulfilled", json.dumps(events[-1])
    assert events[-2]["body"]["outputs"] == {"foo": "banana"}
    assert [event["name"] for event in events] == [
        "vembda.execution.initiated",
        "workflow.execution.initiated",
        "node.execution.initiated",
        "workflow.execution.snapshotted",
        "node.execution.fulfilled",
        "workflow.execution.streaming",
        "workflow.execution.fulfilled",
        "vembda.execution.fulfilled",
    ]


def test_stream_workflow_route__happy_path_run_from_node_with_state(both_stream_types):
    # GIVEN a valid request body representing a run from a node with state
    node_id = uuid4()
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "node_id": str(node_id),
        "environment_api_key": "test",
        "module": "workflow",
        "state": {"meta": {"node_outputs": {"StartNode.Outputs.value": "cherry"}}},
        "files": {
            "__init__.py": "from .display import *",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.start import StartNode
from .nodes.end import EndNode

class Workflow(BaseWorkflow):
    graph = StartNode >> EndNode

    class Outputs(BaseWorkflow.Outputs):
        foo = EndNode.Outputs.value
""",
            "nodes/__init__.py": """
from .start import StartNode
from .end import EndNode

__all__ = ["StartNode", "EndNode"]
""",
            "nodes/start.py": """\
import random
from vellum.workflows.nodes import BaseNode

class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = str
    def run(self) -> Outputs:
        return self.Outputs(value=random.choice(["apple", "banana", "cherry"]))
""",
            "nodes/end.py": """\
from vellum.workflows.nodes import BaseNode
from .start import StartNode

class EndNode(BaseNode):
    fruit = StartNode.Outputs.value.coalesce("date")

    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> Outputs:
        return self.Outputs(value=self.fruit)
""",
            "display/__init__.py": """
from .nodes import *
from .workflow import *
""",
            "display/workflow.py": """
from uuid import UUID
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from vellum_ee.workflows.display.base import EdgeDisplay
from ..nodes.start import StartNode
from ..nodes.end import EndNode

class WorkflowDisplay(BaseWorkflowDisplay):
    edge_displays = {
        (StartNode.Ports.default, EndNode): EdgeDisplay(
            id=UUID("63606ff1-1c70-4516-92c6-cbaba9336424")
        ),
    }
""",
            "display/nodes/__init__.py": """
from .start import StartNodeDisplay
from .end import EndNodeDisplay

__all__ = ["StartNodeDisplay", "EndNodeDisplay"]
""",
            "display/nodes/start.py": """\
from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from ...nodes.start import StartNode

class StartNodeDisplay(BaseNodeDisplay[StartNode]):
    pass
""",
            "display/nodes/end.py": f"""\
from uuid import UUID

from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from ...nodes.end import EndNode

class EndNodeDisplay(BaseNodeDisplay[EndNode]):
    node_id = UUID("{node_id}")
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected workflow and node initiated events
    assert len(events) > 3, json.dumps(events[-1])
    assert events[2]["name"] == "node.execution.initiated", json.dumps(events[-1])
    assert events[2]["body"]["inputs"] == {"fruit": "cherry"}


@mock.patch("workflow_server.api.workflow_view.wait_for_available_process")
def test_stream_workflow_route__concurrent_request_rate_exceeded(mock_wait_for_available_process):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "workspace_api_key": "test",
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    # AND wait_for_available_process returns False
    mock_wait_for_available_process.return_value = False

    # WHEN we call the stream route
    status_code, events = flask_stream(request_body)

    # THEN we get a 429 response
    assert status_code == 429, events

    # AND we get a simple JSON error response
    assert len(events) == 1
    assert events[0] == {
        "detail": f"Workflow server concurrent request rate exceeded. Process count: {get_active_process_count()}"
    }


def test_stream_workflow_route__with_environment_variables(both_stream_types):
    # GIVEN a valid request body with environment variables
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "environment_variables": {"TEST_ENV_VAR": "test_value", "ANOTHER_VAR": "another_value"},
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.references import EnvironmentVariableReference

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        env_var_value = EnvironmentVariableReference(name="TEST_ENV_VAR", default="not_found")
        another_var_value = EnvironmentVariableReference(name="ANOTHER_VAR", default="not_found")
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[1]["name"] == "workflow.execution.initiated"
    assert events[2]["name"] == "workflow.execution.fulfilled"

    # AND the environment variables are accessible in the workflow
    outputs = events[2]["body"]["outputs"]
    assert outputs["env_var_value"] == "test_value"
    assert outputs["another_var_value"] == "another_value"


@mock.patch("workflow_server.api.workflow_view.Queue")
def test_stream_workflow_route__queue_get_timeout(mock_queue_class):
    # GIVEN a valid request body
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    # AND the queue.get method raises Empty exception
    mock_queue_instance = mock_queue_class.return_value
    mock_queue_instance.get.side_effect = Empty()

    # WHEN we call the stream route
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code
        response_data = response.get_json()

    # THEN we get a 408 response
    assert status_code == 408

    # AND we get the expected timeout error message
    assert response_data == {"detail": "Request timed out trying to initiate the Workflow"}


@pytest.mark.parametrize("non_process_stream_types", [code_exec_stream, flask_stream_disable_process_wrapper])
def test_stream_workflow_route__vembda_emitting_calls_monitoring_api(non_process_stream_types):
    """
    Tests that the monitoring API is called when vembda emitting is enabled.
    """

    # GIVEN a valid request body with vembda emitting enabled
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "feature_flags": {"vembda-event-emitting-enabled": True},
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }
    emitted_events = []

    def send_events(self, events: list[WorkflowEvent]) -> None:
        for event in events:
            emitted_events.append(event)

    VellumEmitter._send_events = send_events

    # WHEN we call the stream route with mocked monitoring API
    status_code, events = non_process_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND the expected workflow events were emitted
    event_names = [event.name for event in emitted_events]
    assert len(event_names) == 2, "Should include 2 events"
    assert "workflow.execution.initiated" in event_names, "Should include workflow.execution.initiated event"
    assert "workflow.execution.fulfilled" in event_names, "Should include workflow.execution.fulfilled event"


def test_stream_workflow_route__with_invalid_nested_set_graph(both_stream_types):
    """
    Tests that a workflow with an invalid nested set graph structure raises a clear error in the stream response.
    """
    # GIVEN a Flask application and invalid workflow content with nested set graph
    span_id = uuid4()

    invalid_workflow_content = """
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value = "test"

class InvalidWorkflow(BaseWorkflow):
    graph = {TestNode, {TestNode}}

    class Outputs(BaseWorkflow.Outputs):
        result = TestNode.Outputs.value
"""

    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": invalid_workflow_content,
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events: vembda initiated, workflow initiated, workflow rejected, vembda fulfilled
    assert len(events) == 4

    # AND the first event should be vembda execution initiated
    assert events[0]["name"] == "vembda.execution.initiated"
    assert events[0]["span_id"] == str(span_id)

    # AND the second event should be workflow execution initiated
    assert events[1]["name"] == "workflow.execution.initiated"

    # AND the third event should be workflow execution rejected
    assert events[2]["name"] == "workflow.execution.rejected"
    assert events[1]["span_id"] == events[2]["span_id"]

    # AND the error message should contain information about the invalid graph structure
    error_message = events[2]["body"]["error"]["message"]
    expected_message = (
        "Invalid graph structure detected. "
        "Nested sets or unsupported graph types are not allowed. "
        "Please contact Vellum support for assistance with Workflow configuration."
    )
    assert error_message == expected_message

    # AND the fourth event should be vembda execution fulfilled
    assert events[3]["name"] == "vembda.execution.fulfilled"
    assert events[3]["span_id"] == str(span_id)
    assert events[3]["body"]["exit_code"] == 0


@mock.patch("workflow_server.api.workflow_view.get_is_oom_killed")
def test_stream_workflow_route__oom_does_not_set_timed_out_flag(mock_get_is_oom_killed):
    """
    Tests that when an OOM error occurs, we don't set the timed_out flag in the vembda fulfilled event.
    """
    # GIVEN a workflow that takes some time to execute
    span_id = uuid4()
    request_body = {
        "timeout": 10,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
import time

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.workflows.base import BaseWorkflow


class SlowNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        value: str

    def run(self) -> Outputs:
        time.sleep(2)
        return self.Outputs(value="hello world")


class OOMWorkflow(BaseWorkflow):
    graph = SlowNode
    class Outputs(BaseWorkflow.Outputs):
        final_value = SlowNode.Outputs.value

""",
        },
    }

    # WHEN we mock the OOM killer to trigger after a few checks
    call_count = [0]

    def mock_oom_side_effect():
        call_count[0] += 1
        if call_count[0] > 3:
            return True
        return False

    mock_get_is_oom_killed.side_effect = mock_oom_side_effect

    # AND we call the stream route
    status_code, events = flask_stream(request_body)

    # THEN we get a 200 response
    assert status_code == 200

    # AND we get the expected events
    event_names = [e["name"] for e in events]

    assert "vembda.execution.initiated" in event_names

    # THEN the key assertion: if there's a vembda.execution.fulfilled event, it should NOT have timed_out=True
    vembda_fulfilled_event = next(e for e in events if e["name"] == "vembda.execution.fulfilled")
    assert (
        vembda_fulfilled_event["body"].get("timed_out") is not True
    ), "timed_out flag should not be set when OOM occurs"


@mock.patch("workflow_server.api.workflow_view.ENABLE_PROCESS_WRAPPER", False)
def test_stream_workflow_route__client_disconnect_emits_rejected_event():
    """
    Tests that when a client disconnects mid-stream (GeneratorExit), we emit a workflow execution
    rejected event to the events.create API.
    """
    # GIVEN a valid request body for a workflow that yields multiple events
    span_id = uuid4()
    trace_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "execution_context": {
            "trace_id": str(trace_id),
        },
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    # AND a mock to capture events.create calls
    events_create_calls = []

    def mock_events_create(request):
        events_create_calls.append(request)

    # WHEN we call the stream route and simulate a client disconnect
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        with mock.patch("workflow_server.core.workflow_executor_context.create_vellum_client") as mock_create_client:
            mock_client = mock.MagicMock()
            mock_client.events.create = mock_events_create
            mock_create_client.return_value = mock_client

            response = test_client.post("/workflow/stream", json=request_body)

            # Get the response iterator and consume a few chunks to start the stream
            response_iter = response.response
            next(response_iter)

            # Close the response to trigger GeneratorExit
            response_iter.close()

    # THEN the events.create API should have been called with rejected event
    assert len(events_create_calls) > 0, "events.create should have been called on client disconnect"

    # AND the call should include a workflow.execution.rejected event (sent as SDK event model)
    last_call = events_create_calls[-1]
    assert isinstance(last_call, list), "events.create should be called with a list"
    assert len(last_call) == 1, "Should have exactly one rejected event"

    rejected_event = last_call[0]
    assert rejected_event.name == "workflow.execution.rejected", "Should be a rejected event"

    # AND the rejected event should have the correct error message
    assert "client disconnected" in rejected_event.body.error.message.lower()

    # AND the rejected event should have a workflow_definition
    # TODO: In the future, we should capture the real workflow_definition from the initiated event.
    # For now, we use BaseWorkflow as a placeholder.
    assert rejected_event.body.workflow_definition is not None, "Should have a workflow_definition"


def test_stream_workflow_route__array_input_string_methods(both_stream_types):
    """
    Tests that array inputs of strings can have string methods called on them.

    This is a regression test for APO-2423 where array inputs of strings were being
    deserialized as VellumValue objects instead of plain strings, causing string
    methods like .upper() to fail.
    """

    # GIVEN a workflow that takes an array of strings and calls .upper() on each item
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [
            {
                "name": "items",
                "type": "ARRAY",
                "value": [
                    {"type": "STRING", "value": "hello"},
                    {"type": "STRING", "value": "world"},
                ],
            },
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from typing import List

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state import BaseState


class Inputs(BaseInputs):
    items: List[str]


class UppercaseNode(BaseNode):
    items = Inputs.items

    class Outputs(BaseNode.Outputs):
        result: List[str]

    def run(self) -> Outputs:
        # This should work if items is a list of strings
        # but will fail if items is a list of VellumValue objects
        uppercased = [item.upper() for item in self.items]
        return self.Outputs(result=uppercased)


class Workflow(BaseWorkflow[Inputs, BaseState]):
    graph = UppercaseNode

    class Outputs(BaseWorkflow.Outputs):
        result = UppercaseNode.Outputs.result
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND we should get the expected events without errors
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" in event_names
    assert "workflow.execution.initiated" in event_names
    assert "workflow.execution.fulfilled" in event_names

    # AND the workflow should NOT be rejected
    assert "workflow.execution.rejected" not in event_names, (
        f"Workflow was rejected when it should have succeeded. " f"Events: {events}"
    )

    # AND the output should be the uppercased strings
    fulfilled_event = next(e for e in events if e["name"] == "workflow.execution.fulfilled")
    assert fulfilled_event["body"]["outputs"]["result"] == ["HELLO", "WORLD"]


def test_stream_workflow_route__emits_workflow_events_when_exception_during_trigger_init(both_stream_types):
    """
    Tests that when an exception is raised during trigger __init__ (after workflow import
    but before workflow.execution.initiated is emitted), the API response stream still contains
    workflow lifecycle events.
    """
    # GIVEN a request with a workflow that has a trigger with __init__ that raises
    span_id = uuid4()
    trace_id = uuid4()
    parent_span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "execution_context": {
            "trace_id": str(trace_id),
            "parent_context": {"span_id": str(parent_span_id)},
        },
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "metadata.json": """\
{
    "trigger_path_to_id_mapping": {
        ".triggers.failing_trigger.FailingTrigger": "11111111-1111-1111-1111-111111111111"
    }
}
""",
            "triggers/__init__.py": "",
            "triggers/failing_trigger.py": """\
from vellum.workflows.triggers import BaseTrigger


class FailingTrigger(BaseTrigger):
    pass
""",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow, BaseNode, BaseInputs
from vellum.workflows.triggers import BaseTrigger

from .triggers.failing_trigger import FailingTrigger


class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        result: str

    def run(self):
        return self.Outputs(result="done")


class Workflow(BaseWorkflow):
    graph = FailingTrigger >> StartNode

    class Outputs(BaseWorkflow.Outputs):
        result = StartNode.Outputs.result

    @staticmethod
    def deserialize_trigger(trigger_id: str, inputs: BaseInputs) -> BaseTrigger:
        raise RuntimeError("Exception during trigger deserialization")
""",
        },
        # The trigger_id matches the ID in metadata.json for FailingTrigger
        "trigger_id": "11111111-1111-1111-1111-111111111111",
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND the response should contain vembda events
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" in event_names
    assert "vembda.execution.fulfilled" in event_names

    # AND the response should contain workflow lifecycle events
    assert "workflow.execution.initiated" in event_names, (
        f"Expected workflow.execution.initiated to be emitted when exception occurs during trigger init. "
        f"Events: {event_names}"
    )
    assert "workflow.execution.rejected" in event_names, (
        f"Expected workflow.execution.rejected to be emitted when exception occurs during trigger init. "
        f"Events: {event_names}"
    )
    assert events[-2]["body"]["error"]["code"] == "INTERNAL_ERROR", events[-2]["body"]["error"]["message"]


@pytest.mark.parametrize(
    "event_filter,expected_node_events",
    [
        (None, True),  # Default (EXCLUDE_NESTED_SNAPSHOTTED) includes node events
        ("EXCLUDE_NESTED_SNAPSHOTTED", True),  # Explicitly set includes node events
        ("ALL", True),  # ALL includes node events
        ("ROOT_WORKFLOW_AND_ROOT_NODE", True),  # Includes node events from root workflow
        ("ROOT_WORKFLOW_ONLY", False),  # Only workflow events, no node events
        ("INVALID_FILTER", True),  # Invalid filter falls back to default (includes node events)
    ],
)
def test_stream_workflow_route__event_filter_parameter(both_stream_types, event_filter, expected_node_events):
    """
    Tests that the event_filter parameter controls which events are included in the stream.

    - ROOT_WORKFLOW_ONLY: Only workflow events (initiated, fulfilled, rejected, etc.)
    - ROOT_WORKFLOW_AND_ROOT_NODE: Workflow events + node events from the root workflow
    - EXCLUDE_NESTED_SNAPSHOTTED: All events except snapshotted from nested workflows (default)
    - ALL: All events
    - Invalid filters fall back to default behavior (EXCLUDE_NESTED_SNAPSHOTTED)
    """
    # GIVEN a workflow with a node that will emit node events
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode

class MyNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        result = "hello"

class Workflow(BaseWorkflow):
    graph = MyNode

    class Outputs(BaseWorkflow.Outputs):
        result = MyNode.Outputs.result
""",
        },
    }

    # AND the event_filter is set (or not set for default behavior)
    if event_filter is not None:
        request_body["event_filter"] = event_filter

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND we always get vembda events
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" in event_names
    assert "vembda.execution.fulfilled" in event_names

    # AND we always get workflow events
    assert "workflow.execution.initiated" in event_names
    assert "workflow.execution.fulfilled" in event_names

    # AND node events are included or excluded based on the filter
    has_node_events = any(name.startswith("node.") for name in event_names)
    if expected_node_events:
        assert has_node_events, (
            f"Expected node events to be included with event_filter={event_filter}, " f"but got events: {event_names}"
        )
        assert "node.execution.initiated" in event_names
        assert "node.execution.fulfilled" in event_names
    else:
        assert not has_node_events, (
            f"Expected no node events with event_filter={event_filter}, " f"but got events: {event_names}"
        )


def test_stream_workflow_route__exclude_vembda_events(both_stream_types):
    """
    Tests that when exclude_vembda_events is True, vembda.execution.* events are filtered out.
    """
    # GIVEN a simple workflow
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    # AND exclude_vembda_events is set to True
    request_body["exclude_vembda_events"] = True

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN vembda.execution.* events should be filtered out
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" not in event_names
    assert "vembda.execution.fulfilled" not in event_names

    # AND workflow events should still be present
    assert "workflow.execution.initiated" in event_names
    assert "workflow.execution.fulfilled" in event_names


def test_stream_workflow_route__exclude_vembda_events_workflow_exception(both_stream_types):
    """
    Tests that when exclude_vembda_events is True and an exception occurs during workflow streaming,
    a workflow.execution.rejected event is emitted instead of vembda.execution.fulfilled.
    """
    # GIVEN a workflow that raises an exception during trigger deserialization
    span_id = uuid4()
    trace_id = uuid4()
    parent_span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "execution_context": {
            "trace_id": str(trace_id),
            "parent_context": {"span_id": str(parent_span_id)},
        },
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "metadata.json": """\
{
    "trigger_path_to_id_mapping": {
        ".triggers.failing_trigger.FailingTrigger": "11111111-1111-1111-1111-111111111111"
    }
}
""",
            "triggers/__init__.py": "",
            "triggers/failing_trigger.py": """\
from vellum.workflows.triggers import BaseTrigger


class FailingTrigger(BaseTrigger):
    pass
""",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow, BaseNode, BaseInputs
from vellum.workflows.triggers import BaseTrigger

from .triggers.failing_trigger import FailingTrigger


class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        result: str

    def run(self):
        return self.Outputs(result="done")


class Workflow(BaseWorkflow):
    graph = FailingTrigger >> StartNode

    class Outputs(BaseWorkflow.Outputs):
        result = StartNode.Outputs.result

    @staticmethod
    def deserialize_trigger(trigger_id: str, inputs: BaseInputs) -> BaseTrigger:
        raise RuntimeError("Exception during trigger deserialization")
""",
        },
        "trigger_id": "11111111-1111-1111-1111-111111111111",
    }

    # AND exclude_vembda_events is set to True
    request_body["exclude_vembda_events"] = True

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN vembda.execution.* events should be filtered out
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" not in event_names
    assert "vembda.execution.fulfilled" not in event_names

    # AND workflow.execution.rejected should be emitted
    assert "workflow.execution.initiated" in event_names
    assert "workflow.execution.rejected" in event_names

    # AND the rejected event should have the correct error code
    rejected_event = next(e for e in events if e["name"] == "workflow.execution.rejected")
    assert rejected_event["body"]["error"]["code"] == "INTERNAL_ERROR"


def test_stream_workflow_route__exclude_vembda_events_node_exception(both_stream_types):
    """
    Tests that when exclude_vembda_events is True and an exception occurs during node streaming,
    a node.execution.rejected event is emitted instead of vembda.execution.fulfilled.
    """
    # GIVEN a workflow with a node that raises an exception during execution
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode


class FailingNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        result: str

    def run(self):
        raise RuntimeError("Node execution failed")


class Workflow(BaseWorkflow):
    graph = FailingNode

    class Outputs(BaseWorkflow.Outputs):
        result = FailingNode.Outputs.result
""",
        },
    }

    # AND exclude_vembda_events is set to True
    request_body["exclude_vembda_events"] = True

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN vembda.execution.* events should be filtered out
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" not in event_names
    assert "vembda.execution.fulfilled" not in event_names

    # AND node.execution.rejected should be emitted (the node itself rejects)
    assert "node.execution.rejected" in event_names

    # AND workflow.execution.rejected should also be emitted (workflow rejects due to node failure)
    assert "workflow.execution.rejected" in event_names


def test_stream_workflow_route__generator_in_node_output_skips_failed_event_and_continues(both_stream_types):
    """
    Tests that when a node output contains a generator object that causes a PydanticSerializationError,
    the failed event is skipped and the workflow continues to completion.
    """
    # GIVEN a workflow where a node outputs a generator object
    span_id = uuid4()
    trace_id = uuid4()
    parent_span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "execution_context": {
            "trace_id": str(trace_id),
            "parent_context": {"span_id": str(parent_span_id)},
        },
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from typing import Any, Iterator
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode


class GeneratorNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        data: Any

    def run(self):
        def my_generator():
            yield 1
            yield 2
            yield 3
        return self.Outputs(data=my_generator())


class Workflow(BaseWorkflow):
    graph = GeneratorNode

    class Outputs(BaseWorkflow.Outputs):
        result = GeneratorNode.Outputs.data
""",
        },
    }

    # WHEN we call the stream route
    status_code, events = both_stream_types(request_body)

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND we should have at least a vembda.execution.initiated event
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" in event_names

    # AND the workflow should complete (either fulfilled or rejected based on workflow logic)
    # The key is that the serialization error doesn't kill the workflow - it just skips the failed event
    assert (
        "workflow.execution.initiated" in event_names
    ), f"Expected workflow.execution.initiated event, got: {event_names}"

    # AND we should have a terminal workflow event (fulfilled or rejected)
    has_terminal_event = "workflow.execution.fulfilled" in event_names or "workflow.execution.rejected" in event_names
    assert has_terminal_event, f"Expected a terminal workflow event (fulfilled or rejected), got: {event_names}"


@mock.patch("workflow_server.api.workflow_view.ENABLE_PROCESS_WRAPPER", False)
def test_stream_workflow_route__serialization_error_skips_event_and_continues():
    """
    Tests that when a PydanticSerializationError occurs during event serialization,
    the failed event is skipped and the workflow continues to completion.
    """
    # GIVEN a valid workflow with a specified workflow_span_id
    workflow_span_id = uuid4()
    trace_id = uuid4()
    parent_span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(uuid4()),
        "workflow_span_id": str(workflow_span_id),
        "execution_context": {
            "trace_id": str(trace_id),
            "parent_context": {"span_id": str(parent_span_id)},
        },
        "inputs": [],
        "environment_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow


class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        result = "done"
""",
        },
    }

    # AND _dump_event is mocked to raise a PydanticSerializationError on the second call
    # (first call is for vembda.execution.initiated, second is for workflow.execution.initiated)
    # Note: We use the _dump_event imported at the top of this file (before any patching)
    call_count = [0]

    def mock_dump_event(event, executor_context):
        call_count[0] += 1
        if call_count[0] == 2:
            raise PydanticSerializationError("Object of type generator is not JSON serializable")
        return _dump_event(event, executor_context)

    # WHEN we call the stream route with the mocked _dump_event
    flask_app = create_app()
    with flask_app.test_client() as test_client:
        with mock.patch("workflow_server.core.executor._dump_event", side_effect=mock_dump_event):
            response = test_client.post("/workflow/stream", json=request_body)
            status_code = response.status_code

            events = [
                json.loads(line)
                for line in response.data.decode().split("\n")
                if line and line not in ["WAITING", "END"]
            ]

    # THEN we get a 200 response
    assert status_code == 200, events

    # AND we should have at least a vembda.execution.initiated event
    event_names = [e["name"] for e in events]
    assert "vembda.execution.initiated" in event_names

    # AND the workflow should complete with a vembda.execution.fulfilled event
    # (the workflow.execution.fulfilled event may have been skipped due to serialization error)
    assert "vembda.execution.fulfilled" in event_names, f"Expected vembda.execution.fulfilled event, got: {event_names}"

    # AND the workflow.execution.initiated event should still be present (it was the first event, before the error)
    assert (
        "workflow.execution.initiated" in event_names
    ), f"Expected workflow.execution.initiated event, got: {event_names}"



from datetime import datetime, timezone
from io import StringIO
import logging
from multiprocessing import Process, Queue
import os
import random
import string
import sys
from threading import Event as ThreadingEvent
import time
from traceback import format_exc
from uuid import UUID, uuid4
from typing import Any, Callable, Generator, Iterator, Optional, Tuple

import orjson
from pydantic_core import PydanticSerializationError
from vellum_ee.workflows.display.utils.events import event_enricher
from vellum_ee.workflows.display.utils.expressions import base_descriptor_validator
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder

from vellum.workflows import BaseWorkflow
from vellum.workflows.context import execution_context
from vellum.workflows.emitters.base import BaseWorkflowEmitter
from vellum.workflows.emitters.vellum_emitter import VellumEmitter
from vellum.workflows.errors import WorkflowError, WorkflowErrorCode
from vellum.workflows.events.exception_handling import stream_initialization_exception
from vellum.workflows.events.node import NodeExecutionRejectedBody, NodeExecutionRejectedEvent
from vellum.workflows.events.types import BaseEvent
from vellum.workflows.events.workflow import (
    WorkflowEvent,
    WorkflowExecutionInitiatedBody,
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionRejectedBody,
    WorkflowExecutionRejectedEvent,
)
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode
from vellum.workflows.nodes.mocks import MockNodeExecution
from vellum.workflows.resolvers.base import BaseWorkflowResolver
from vellum.workflows.resolvers.resolver import VellumResolver
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.state.store import EmptyStore
from vellum.workflows.triggers import BaseTrigger
from vellum.workflows.types import CancelSignal
from vellum.workflows.workflows.event_filters import (
    all_workflow_event_filter,
    root_workflow_event_filter,
    workflow_event_filter,
    workflow_sandbox_event_filter,
)
from workflow_server.config import LOCAL_DEPLOYMENT, LOCAL_WORKFLOW_MODULE
from workflow_server.core.cancel_workflow import CancelWorkflowWatcherThread
from workflow_server.core.events import (
    SPAN_ID_EVENT,
    STREAM_FINISHED_EVENT,
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
)
from workflow_server.core.utils import (
    create_vembda_rejected_event,
    is_events_emitting_enabled,
    serialize_vembda_rejected_event,
)
from workflow_server.core.workflow_executor_context import (
    BaseExecutorContext,
    EventFilterType,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.log_proxy import redirect_log
from workflow_server.utils.system_utils import get_memory_in_use_mb
from workflow_server.utils.utils import get_version

logger = logging.getLogger(__name__)

EVENT_FILTER_MAP: dict[EventFilterType, Callable[[type["BaseWorkflow"], "WorkflowEvent"], bool]] = {
    "ROOT_WORKFLOW_ONLY": workflow_event_filter,
    "ROOT_WORKFLOW_AND_ROOT_NODE": root_workflow_event_filter,
    "EXCLUDE_NESTED_SNAPSHOTTED": workflow_sandbox_event_filter,
    "ALL": all_workflow_event_filter,
}


def _get_event_filter(
    filter_type: Optional[EventFilterType],
) -> Callable[[type["BaseWorkflow"], "WorkflowEvent"], bool]:
    """
    Get the event filter function based on the filter type.
    Defaults to workflow_sandbox_event_filter if no filter type is specified.
    """
    if filter_type is None:
        return workflow_sandbox_event_filter
    return EVENT_FILTER_MAP.get(filter_type, workflow_sandbox_event_filter)


def stream_node_process_timeout(
    executor_context: NodeExecutorContext,
    queue: Queue,
) -> Process:
    node_process = Process(
        target=_stream_node_wrapper,
        args=(executor_context, queue),
    )
    node_process.start()

    if node_process.exitcode is not None:
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error", timed_out=True))

    return node_process


def _stream_node_wrapper(executor_context: NodeExecutorContext, queue: Queue) -> None:
    try:
        for event in stream_node(executor_context=executor_context):
            queue.put(event)
    except WorkflowInitializationException as e:
        queue.put(create_vembda_rejected_event(executor_context, e.message))
    except Exception as e:
        logger.exception(e)
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error"))


def _stream_workflow_wrapper(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
    cancel_signal: CancelSignal,
    timeout_signal: CancelSignal,
) -> None:
    span_id_emitted = False
    try:
        stream_iterator, span_id = stream_workflow(
            executor_context=executor_context,
            cancel_signal=cancel_signal,
            timeout_signal=timeout_signal,
        )

        queue.put(f"{SPAN_ID_EVENT}:{span_id}")
        span_id_emitted = True

        for event in stream_iterator:
            queue.put(orjson.dumps(event).decode("utf-8"))

    except Exception as e:
        if not span_id_emitted:
            queue.put(f"{SPAN_ID_EVENT}:{uuid4()}")

        logger.exception(e)
        queue.put(serialize_vembda_rejected_event(executor_context, "Internal Server Error"))

    queue.put(STREAM_FINISHED_EVENT)

    exit(0)


def stream_workflow_process_timeout(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
    cancel_signal: CancelSignal,
    timeout_signal: CancelSignal,
) -> Process:
    workflow_process = Process(
        target=_stream_workflow_wrapper,
        args=(
            executor_context,
            queue,
            cancel_signal,
            timeout_signal,
        ),
    )
    workflow_process.start()

    if workflow_process.exitcode is not None:
        queue.put(create_vembda_rejected_event(executor_context, "Internal Server Error", timed_out=True))

    return workflow_process


def stream_workflow(
    executor_context: WorkflowExecutorContext,
    timeout_signal: CancelSignal,
    cancel_signal: CancelSignal,
    disable_redirect: bool = True,
) -> tuple[Iterator[dict], UUID]:
    cancel_watcher_kill_switch = ThreadingEvent()
    workflow: Optional[BaseWorkflow] = None
    try:
        workflow, namespace = _create_workflow(executor_context)

        trigger_id = executor_context.trigger_id

        inputs_or_trigger = workflow.deserialize_trigger(
            trigger_id=trigger_id,
            inputs=executor_context.inputs,
            dataset_row=executor_context.dataset_row,
        )

        # Determine whether we have inputs or a trigger
        if isinstance(inputs_or_trigger, BaseInputs):
            workflow_inputs = inputs_or_trigger
            trigger = None
        elif isinstance(inputs_or_trigger, BaseTrigger):
            workflow_inputs = None
            trigger = inputs_or_trigger
        else:
            workflow_inputs = None
            trigger = None

        workflow_state = (
            workflow.deserialize_state(
                executor_context.state,
                workflow_inputs=workflow_inputs or BaseInputs(),
            )
            if executor_context.state
            else None
        )
        node_output_mocks = MockNodeExecution.validate_all(
            executor_context.node_output_mocks,
            workflow.__class__,
            descriptor_validator=base_descriptor_validator,
        )

        cancel_signal = cancel_signal or ThreadingEvent()

        stream = workflow.stream(
            inputs=workflow_inputs,
            state=workflow_state,
            node_output_mocks=node_output_mocks,
            event_filter=_get_event_filter(executor_context.event_filter),
            cancel_signal=cancel_signal,
            entrypoint_nodes=[executor_context.node_id] if executor_context.node_id else None,
            previous_execution_id=executor_context.previous_execution_id,
            timeout=executor_context.timeout,
            trigger=trigger,
            execution_id=executor_context.workflow_span_id,
            event_max_size=executor_context.event_max_size,
        )
    except WorkflowInitializationException as e:
        cancel_watcher_kill_switch.set()

        with execution_context(
            parent_context=executor_context.execution_context.parent_context,
            trace_id=executor_context.execution_context.trace_id,
        ):
            initialization_exception_stream = stream_initialization_exception(e)

        def _stream_generator() -> Generator[dict[str, Any], Any, None]:
            for event in initialization_exception_stream:
                if workflow:
                    for emitter in workflow.emitters:
                        emitter.emit_event(event)

                yield _dump_event(
                    event=event,
                    executor_context=executor_context,
                )

            if workflow:
                workflow.join()

        return (
            _call_stream(
                executor_context=executor_context,
                stream_generator=_stream_generator,
                disable_redirect=disable_redirect,
                timeout_signal=timeout_signal,
            ),
            initialization_exception_stream.span_id,
        )
    except Exception:
        cancel_watcher_kill_switch.set()
        logger.exception("Failed to generate Workflow Stream")

        parent = executor_context.execution_context.parent_context
        trace_id = executor_context.execution_context.trace_id
        span_id = uuid4()
        workflow_definition = workflow.__class__ if workflow is not None else BaseWorkflow

        def _stream_generator() -> Generator[dict[str, Any], Any, None]:
            initiated_event: WorkflowEvent = WorkflowExecutionInitiatedEvent(
                trace_id=trace_id,
                span_id=span_id,
                body=WorkflowExecutionInitiatedBody(
                    workflow_definition=workflow_definition,
                    inputs=BaseInputs(),
                    initial_state=None,
                ),
                parent=parent,
            )
            yield _dump_event(
                event=initiated_event,
                executor_context=executor_context,
            )

            rejected_event = WorkflowExecutionRejectedEvent(
                trace_id=trace_id,
                span_id=span_id,
                body=WorkflowExecutionRejectedBody(
                    workflow_definition=workflow_definition,
                    error=WorkflowError(
                        code=WorkflowErrorCode.INTERNAL_ERROR,
                        message="Failed to generate Workflow Stream",
                    ),
                    stacktrace=format_exc(),
                ),
                parent=parent,
            )
            yield _dump_event(
                event=rejected_event,
                executor_context=executor_context,
            )

        return (
            _call_stream(
                executor_context=executor_context,
                stream_generator=_stream_generator,
                disable_redirect=disable_redirect,
                timeout_signal=timeout_signal,
            ),
            span_id,
        )

    cancel_watcher = CancelWorkflowWatcherThread(
        kill_switch=cancel_watcher_kill_switch,
        execution_id=stream.span_id,
        timeout_seconds=executor_context.timeout,
        vembda_public_url=executor_context.vembda_public_url,
        cancel_signal=cancel_signal,
    )

    try:
        if executor_context.vembda_public_url:
            cancel_watcher.start()
    except Exception:
        logger.exception("Failed to start cancel watcher")

    def call_workflow() -> Generator[dict[str, Any], Any, None]:
        try:
            first = True
            for event in stream:
                if first:
                    executor_context.stream_start_time = time.time_ns()
                    first = False

                try:
                    yield _dump_event(
                        event=event,
                        executor_context=executor_context,
                    )
                except PydanticSerializationError:
                    logger.exception(
                        "Failed to serialize event from Workflow Stream",
                        extra={"event_name": event.name, "event_span_id": str(event.span_id)},
                    )
                    # Skip this event and continue processing the rest of the stream
        except Exception as e:
            logger.exception("Failed to generate event from Workflow Stream")
            raise e
        finally:
            cancel_watcher_kill_switch.set()

        if workflow is not None:
            workflow.join()

    return (
        _call_stream(
            executor_context=executor_context,
            stream_generator=call_workflow,
            disable_redirect=disable_redirect,
            timeout_signal=timeout_signal,
        ),
        stream.span_id,
    )


def stream_node(
    executor_context: NodeExecutorContext,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    workflow, namespace = _create_workflow(executor_context)

    def call_node() -> Generator[dict[str, Any], Any, None]:
        executor_context.stream_start_time = time.time_ns()

        for event in workflow.run_node(executor_context.node_ref, inputs=executor_context.inputs):
            yield event.model_dump(mode="json")

    return _call_stream(
        executor_context=executor_context,
        stream_generator=call_node,
        disable_redirect=disable_redirect,
        timeout_signal=ThreadingEvent(),
    )


def _call_stream(
    executor_context: BaseExecutorContext,
    stream_generator: Callable[[], Generator[dict[str, Any], Any, None]],
    timeout_signal: CancelSignal,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    log_redirect: Optional[StringIO] = None

    if not disable_redirect:
        log_redirect = redirect_log()

    try:
        yield from stream_generator()

        if not executor_context.exclude_vembda_events:
            vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=0,
                    log=log_redirect.getvalue() if log_redirect else "",
                    stderr="",
                    container_overhead_latency=executor_context.container_overhead_latency,
                    timed_out=timeout_signal.is_set(),
                ),
                parent=None,
            )
            yield vembda_fulfilled_event.model_dump(mode="json")

    except Exception:
        if not executor_context.exclude_vembda_events:
            vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    log=log_redirect.getvalue() if log_redirect else "",
                    stderr=format_exc(),
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            )
            yield vembda_fulfilled_event.model_dump(mode="json")
        else:
            parent = executor_context.execution_context.parent_context if executor_context.execution_context else None
            if isinstance(executor_context, NodeExecutorContext):
                rejected_event: BaseEvent = NodeExecutionRejectedEvent(
                    trace_id=executor_context.trace_id,
                    span_id=executor_context.execution_id,
                    body=NodeExecutionRejectedBody(
                        node_definition=BaseNode,
                        error=WorkflowError(
                            code=WorkflowErrorCode.INTERNAL_ERROR,
                            message="Node execution failed",
                        ),
                        stacktrace=format_exc(),
                    ),
                    parent=parent,
                )
            else:
                rejected_event = WorkflowExecutionRejectedEvent(
                    trace_id=executor_context.trace_id,
                    span_id=executor_context.execution_id,
                    body=WorkflowExecutionRejectedBody(
                        workflow_definition=BaseWorkflow,
                        error=WorkflowError(
                            code=WorkflowErrorCode.INTERNAL_ERROR,
                            message="Workflow execution failed",
                        ),
                        stacktrace=format_exc(),
                    ),
                    parent=parent,
                )
            yield _dump_event(
                event=rejected_event,
                executor_context=executor_context,
            )


def _create_workflow(executor_context: BaseExecutorContext) -> Tuple[BaseWorkflow, str]:
    namespace = _get_file_namespace(executor_context)
    if namespace != LOCAL_WORKFLOW_MODULE:
        sys.meta_path.append(
            VirtualFileFinder(executor_context.files, namespace, source_module=executor_context.module)
        )

    workflow_context = _create_workflow_context(executor_context)
    Workflow = BaseWorkflow.load_from_module(namespace)
    VembdaExecutionFulfilledEvent.model_rebuild(
        # Not sure why this is needed, but it is required for the VembdaExecutionFulfilledEvent to be
        # properly rebuilt with the recursive types.
        # use flag here to determine which emitter to use
        _types_namespace={
            "BaseWorkflow": BaseWorkflow,
            "BaseNode": BaseNode,
        },
    )

    # Determine whether to enable the Vellum Emitter for event publishing
    use_vellum_emitter = is_events_emitting_enabled(executor_context)
    emitters: list["BaseWorkflowEmitter"] = []
    if use_vellum_emitter:
        emitters = [VellumEmitter()]

    use_vellum_resolver = executor_context.previous_execution_id is not None
    resolvers: list["BaseWorkflowResolver"] = []
    if use_vellum_resolver:
        resolvers = [VellumResolver()]

    # Explicit constructor call to satisfy typing
    workflow = Workflow(
        context=workflow_context,
        store=EmptyStore(),
        emitters=emitters,
        resolvers=resolvers,
    )

    return workflow, namespace


def _create_workflow_context(executor_context: BaseExecutorContext) -> WorkflowContext:
    if executor_context.environment_variables:
        os.environ.update(executor_context.environment_variables)

    namespace = _get_file_namespace(executor_context)

    return WorkflowContext(
        vellum_client=executor_context.vellum_client,
        execution_context=executor_context.execution_context,
        generated_files=executor_context.files,
        namespace=namespace,
    )


def _get_file_namespace(executor_context: BaseExecutorContext) -> str:
    if (
        LOCAL_WORKFLOW_MODULE
        and hasattr(executor_context.execution_context.parent_context, "deployment_name")
        and LOCAL_DEPLOYMENT == executor_context.execution_context.parent_context.deployment_name  # type: ignore
    ):
        return LOCAL_WORKFLOW_MODULE

    if executor_context.execution_id:
        return str(executor_context.execution_id)

    return get_random_namespace()


def get_random_namespace() -> str:
    return "workflow_tmp_" + "".join(random.choice(string.ascii_letters + string.digits) for i in range(14))


def _enrich_event(event: WorkflowEvent, executor_context: Optional[BaseExecutorContext] = None) -> WorkflowEvent:
    """
    Enrich an event with metadata based on the event type.

    For initiated events, include server and SDK versions.
    For fulfilled events with WORKFLOW_DEPLOYMENT parent, include memory usage.
    """
    metadata: Optional[dict] = None

    try:
        is_deployment = event.parent and event.parent.type in ("WORKFLOW_DEPLOYMENT", "EXTERNAL")

        if event.name == "workflow.execution.initiated" and is_deployment:
            metadata = {
                **get_version(),
            }

            memory_mb = get_memory_in_use_mb()
            if memory_mb is not None:
                metadata["memory_usage_mb"] = memory_mb

            if executor_context is not None:
                metadata["is_new_server"] = executor_context.is_new_server

                if executor_context.vembda_service_initiated_timestamp is not None and event.timestamp is not None:
                    event_ts = event.timestamp
                    if event_ts.tzinfo is None:
                        event_ts = event_ts.replace(tzinfo=timezone.utc)
                    event_ts_ns = int(event_ts.timestamp() * 1_000_000_000)
                    initiated_latency = event_ts_ns - executor_context.vembda_service_initiated_timestamp
                    metadata["initiated_latency"] = initiated_latency
        elif event.name == "workflow.execution.fulfilled" and is_deployment:
            metadata = {}
            memory_mb = get_memory_in_use_mb()
            if memory_mb is not None:
                metadata["memory_usage_mb"] = memory_mb
    except Exception:
        pass

    vellum_client = executor_context.vellum_client if executor_context else None
    return event_enricher(event, vellum_client, metadata=metadata)


def _dump_event(event: BaseEvent, executor_context: BaseExecutorContext) -> dict:
    module_base = executor_context.module.split(".")

    dump = event.model_dump(
        mode="json",
        context={"event_enricher": lambda event: _enrich_event(event, executor_context)},
    )
    if dump["name"] in {
        "workflow.execution.initiated",
        "workflow.execution.fulfilled",
        "workflow.execution.rejected",
        "workflow.execution.streaming",
        "workflow.execution.paused",
        "workflow.execution.resumed",
    }:
        dump["body"]["workflow_definition"]["module"] = module_base + dump["body"]["workflow_definition"]["module"][1:]
    elif dump["name"] in {
        "node.execution.initiated",
        "node.execution.fulfilled",
        "node.execution.rejected",
        "node.execution.streaming",
        "node.execution.paused",
        "node.execution.resumed",
    }:
        dump["body"]["node_definition"]["module"] = module_base + dump["body"]["node_definition"]["module"][1:]

    return dump



from dataclasses import field
from functools import cached_property
import os
import time
from uuid import UUID
from typing import Any, Literal, Optional, Union
from typing_extensions import Self

from flask import has_request_context, request
from pydantic import Field, field_validator, model_validator

from vellum import ApiVersionEnum, Vellum
from vellum.client.core import UniversalBaseModel
from vellum.workflows.context import ExecutionContext
from vellum.workflows.vellum_client import create_vellum_client
from workflow_server.utils.utils import convert_json_inputs_to_vellum

DEFAULT_TIMEOUT_SECONDS = int(os.getenv("MAX_WORKFLOW_RUNTIME_SECONDS", 1800))

EventFilterType = Literal["ROOT_WORKFLOW_ONLY", "ROOT_WORKFLOW_AND_ROOT_NODE", "EXCLUDE_NESTED_SNAPSHOTTED", "ALL"]
VALID_EVENT_FILTER_VALUES = {"ROOT_WORKFLOW_ONLY", "ROOT_WORKFLOW_AND_ROOT_NODE", "EXCLUDE_NESTED_SNAPSHOTTED", "ALL"}


class BaseExecutorContext(UniversalBaseModel):
    inputs: dict = Field(default_factory=dict)
    state: Optional[dict] = None
    timeout: int = DEFAULT_TIMEOUT_SECONDS
    files: dict[str, str]
    environment_api_key: str
    api_version: Optional[ApiVersionEnum] = None
    execution_id: UUID
    module: str
    execution_context: ExecutionContext = field(default_factory=ExecutionContext)
    request_start_time: int = Field(default_factory=lambda: time.time_ns())
    stream_start_time: int = 0
    vembda_public_url: Optional[str] = None
    node_output_mocks: Optional[list[Any]] = None
    environment_variables: Optional[dict[str, str]] = None
    previous_execution_id: Optional[UUID] = None
    feature_flags: Optional[dict[str, bool]] = None
    is_new_server: bool = False
    trigger_id: Optional[Union[UUID, str]] = None
    dataset_row: Optional[Union[int, str, UUID]] = None
    # The actual 'execution id' of the workflow that we pass into the workflow
    # when running in async mode.
    workflow_span_id: Optional[UUID] = None
    vembda_service_initiated_timestamp: Optional[int] = None
    event_filter: Optional[EventFilterType] = None
    event_max_size: Optional[int] = None
    exclude_vembda_events: bool = False
    predict_api_url: Optional[str] = None

    @field_validator("event_filter", mode="before")
    @classmethod
    def validate_event_filter(cls, v: Any) -> Optional[EventFilterType]:
        """Return None for unexpected filter values instead of raising a validation error."""
        if v is None:
            return None
        if v in VALID_EVENT_FILTER_VALUES:
            return v
        return None

    @field_validator("inputs", mode="before")
    @classmethod
    def convert_inputs(cls, v: Any) -> dict:
        if v is None:
            return {}
        if isinstance(v, list):
            return convert_json_inputs_to_vellum(v)
        return v

    @field_validator("api_version", mode="before")
    @classmethod
    def extract_api_version_from_headers(cls, v: Any) -> Any:
        if v is not None:
            return v
        if has_request_context():
            api_version_header = request.headers.get("x-api-version")
            if api_version_header:
                return api_version_header
        return v

    @property
    def container_overhead_latency(self) -> int:
        return self.stream_start_time - self.request_start_time if self.stream_start_time else -1

    @property
    def trace_id(self) -> UUID:
        return self.execution_context.trace_id

    @cached_property
    def vellum_client(self) -> Vellum:
        return create_vellum_client(
            api_key=self.environment_api_key,
            api_version=self.api_version,
            predict_api_url=self.predict_api_url,
        )

    def __hash__(self) -> int:
        # do we think we need anything else for a unique hash for caching?
        return hash(str(self.execution_id))


class WorkflowExecutorContext(BaseExecutorContext):
    node_id: Optional[UUID] = None  # Sent during run from node UX


class NodeExecutorContext(BaseExecutorContext):
    node_id: Optional[UUID] = None
    node_module: Optional[str] = None
    node_name: Optional[str] = None

    @property
    def node_ref(self) -> Union[UUID, str]:
        """
        Returns the node reference for use with workflow.run_node().

        Returns node_id if it exists, otherwise returns the combination
        of node_module and node_name as a fully qualified string.
        """
        if self.node_id:
            return self.node_id
        return f"{self.node_module}.{self.node_name}"

    @model_validator(mode="after")
    def validate_node_identification(self) -> Self:
        if not self.node_id and not (self.node_module and self.node_name):
            raise ValueError("Either node_id or both node_module and node_name must be provided")
        return self



from functools import cache
import logging
import pathlib
import re

from workflow_server.core.type_checker.base import ConfigurationStatus, TypeCheckResult, WorkflowTypeChecker

logger = logging.getLogger(__name__)


@cache
def _get_mypy_overrides() -> list[str]:
    import importlib.util

    if not importlib.util.find_spec("mypy"):
        return []

    from importlib.metadata import version

    from packaging.version import parse as parse_version

    flags = []
    try:
        if parse_version(version("mypy")) >= parse_version("1.19.0"):
            flags.append("--fixed-format-cache")
    except Exception:
        logger.warning("Failed to determine mypy version. Using default cache format.")

    return flags


class MypyWorkflowTypeChecker(WorkflowTypeChecker):
    def is_configured(self) -> ConfigurationStatus:
        try:
            import mypy.api  # noqa: F401
        except ImportError:
            return False, "mypy is not installed in the current environment"

        self._is_configured = True
        return True, None

    def _run(self, dir: pathlib.Path) -> TypeCheckResult:
        if not self._is_configured:
            return TypeCheckResult(
                success=False,
                failure_message="mypy is not configured",
            )

        import mypy.api

        data_path = pathlib.Path(__file__).parent / "serialization.mypy.ini"
        stdout, stderr, exit_code = mypy.api.run([str(dir), "--config-file", str(data_path), *_get_mypy_overrides()])
        return TypeCheckResult(
            success=exit_code == 0,
            type_errors=stdout,
            failure_message=stderr,
        )

    def _clean_type_errors(self, type_errors: str, dir: pathlib.Path) -> str:
        dir_prefix = re.escape(str(dir))
        pattern = re.compile(f"{dir_prefix}/?")
        return pattern.sub("", type_errors)
